﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: BackgroundSubtraction.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System.Collections.Generic;
using System.ComponentModel;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.OpenGL;

namespace RGBZcamera.Library
{
	public enum MattingTechnique
	{
		BinaryMatte        = 0,
		BlurredBinaryMatte = 1,
		DistanceSoftMatte  = 2,
	};

	/// <summary>TODO</summary>
	public class BackgroundSubtraction : FboNode
	{
		[Uniform, Description("Background plane equation for background subtraction (Xx+Yy+Zy+W=0)."), DisplayName("Background plane"), Category("Background")]
		public Vector4 BackgroundPlane { get; set; }

		[Uniform, DefaultValue(float.NegativeInfinity), Description("Threshold for background segmentation."), DisplayName("Plane threshold"), Category("Background")]
		public float PlaneThreshold { get; set; }

		[Uniform]
		public bool ShowMatte { get; set; }

		[Uniform]
		public Vector3 BackgroundColour { get; set; }

		[Uniform]
		public MattingTechnique MattingTechnique { get; set; }

		public bool ShouldSerializeBackgroundPlane() { return !BackgroundPlane.Equals(new Vector4(0, 0, -1, 1)); }
		public void ResetBackgroundPlane() { BackgroundPlane = new Vector4(0, 0, -1, 1); }

		public bool ShouldSerializeBackgroundColour() { return !BackgroundColour.Equals(new Vector3(0, 0, 0)); }
		public void ResetBackgroundColour() { BackgroundColour = new Vector3(0, 0, 0); }

		/// <summary>Initialises a new instance of the <see cref="BackgroundSubtraction"/> class.</summary>
		public BackgroundSubtraction()
		{
			BackgroundPlane  = new Vector4(0, 0, -1, 1); // default: plane at 1m depth from camera
			PlaneThreshold   = float.NegativeInfinity; // disable thresholding
			ShowMatte        = false;
			BackgroundColour = new Vector3(0, 0, 0);
			MattingTechnique = MattingTechnique.BlurredBinaryMatte;

			AddInputPins("colour", "points");
			AddOutputPin("out");
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.Mesh.vert", ShaderType.FragmentShader),
				Utils.CreateShaderFromResource("Shaders.BackgroundSubtraction.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Name = "colour", Unit = 0 },
				new TextureBinding { Name = "points", Unit = 1 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			AssertFrameProperties(inputs, FrameProps.Size);

			// allocate a texture that is identical to the input
			var resultTexture = TextureFrame.CreateToMatch(inputs["colour"]);
			resultTexture.Alloc();

			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}
	}
}
