﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: ConfigReader.cs 555 2012-09-12 10:11:44Z cr333 $
*/
using System;
using System.ComponentModel;
using System.IO;
using OpenTK;
using VideoLib.OpenGL;

namespace RGBZcamera.Library.Config
{
	/// <summary>Reads in a dataset configuration file.</summary>
	public class ConfigReader
	{
		/// <summary>Gets video camera configuration data.</summary>
		public CameraReader Video  { get; protected set; }

		/// <summary>Gets depth sensor configuration data.</summary>
		public CameraReader Depth { get; protected set; }

		/// <summary>Gets the name of the dataset.</summary>
		public string DatasetName { get; protected set; }

		/// <summary>Gets the input directory.</summary>
		public string InputDir  { get; protected set; }

		/// <summary>Gets the output directory.</summary>
		public string OutputDir { get; protected set; }

		/// <summary>Gets the results directory.</summary>
		public string ResultsDir { get; protected set; }

		private IniFile iniFile;

		/// <summary>Initialises a new instance of the <see cref="ConfigReader"/> class.</summary>
		/// <param name="filename">The filename of the config INI file.</param>
		public ConfigReader(string filename)
		{
			Video   = new CameraReader(filename, "Video");
			Depth   = new CameraReader(filename, "Depth");
			iniFile = new IniFile(filename);

			var dataDir = Path.GetDirectoryName(Path.GetDirectoryName(Path.GetDirectoryName(filename)));
			DatasetName = Path.GetFileName(Path.GetDirectoryName(filename));
			InputDir   = Path.GetDirectoryName(filename);
			OutputDir  = Path.Combine(Path.Combine(dataDir, "output"), DatasetName);
			ResultsDir = Path.Combine(Path.Combine(dataDir, "results"), DatasetName);
		}

		/// <summary>Finds the data directory ("/data/" in some parent directory of the provided directory).</summary>
		/// <returns>Null if the directory is not found.</returns>
		public static DirectoryInfo FindDataDirectory(string startDirectory)
		{
			// Start search in the given directory ...
			var cwd = new DirectoryInfo(startDirectory);

			// ... and go to parent directory until we reach the root (when cwd is null) ...
			while (cwd != null)
			{
				var path = Path.Combine(cwd.FullName, "data");

				// ... or the data directory is found.
				if (Directory.Exists(path))
					return new DirectoryInfo(path);

				cwd = cwd.Parent; // NB: Parent of a root is null
			}

			// Data directory not found.
			return null;
		}

		/// <summary>Reads a setting as a float array from the config file.</summary>
		/// <param name="section">Section in the config file.</param>
		/// <param name="name">Name of the setting.</param>
		/// <param name="defaultValue">The default value, if the setting is not found.</param>
		public float[] ReadSetting(string section, string name, float[] defaultValue)
		{
			// try to read in the corresponding float tuple
			var floats = iniFile.ReadFloatTuple(section, name);

			// check for an empty array -> setting not found
			if (floats.Length == 0)
				return defaultValue;

			return floats;
		}

		/// <summary>Reads a setting as a float from the config file.</summary>
		/// <param name="section">Section in the config file.</param>
		/// <param name="name">Name of the setting.</param>
		/// <param name="defaultValue">The default value, if the setting is not found.</param>
		public float ReadSetting(string section, string name, float defaultValue)
		{
			// try to read in as string first
			var value = iniFile.ReadString(section, name);

			// check for an empty string -> setting not found
			if (value.Length == 0)
				return defaultValue;

			return float.Parse(value.TrimEnd('f'), iniFile.Culture);
		}

		/// <summary>Reads a setting as a int from the config file.</summary>
		/// <param name="section">Section in the config file.</param>
		/// <param name="name">Name of the setting.</param>
		/// <param name="defaultValue">The default value, if the setting is not found.</param>
		public int ReadSetting(string section, string name, int defaultValue)
		{
			// try to read in as string first
			var value = iniFile.ReadString(section, name);

			// check for an empty string -> setting not found
			if (value.Length == 0)
				return defaultValue;

			return int.Parse(value, iniFile.Culture);
		}

		/// <summary>Reads a setting as an enum from the config file.</summary>
		/// <param name="section">Section in the config file.</param>
		/// <param name="name">Name of the setting.</param>
		/// <param name="defaultValue">The default value, if the setting is not found.</param>
		public int ReadSetting(string section, string name, Enum defaultValue)
		{
			// read the setting in as string first
			var value = iniFile.ReadString(section, name);

			// check for an empty string -> setting not found
			if (value.Length == 0)
				return (int)(object)defaultValue;

			// try to parse as enum first
			if (Enum.IsDefined(defaultValue.GetType(), value))
			{
				TypeConverter converter = TypeDescriptor.GetConverter(defaultValue.GetType());
				var val = converter.ConvertFromString(value);
				return (int)val;
			}

			// finally parse as int
			return int.Parse(value, iniFile.Culture);
		}

		/// <summary>Reads a setting as a bool from the config file.</summary>
		/// <param name="section">Section in the config file.</param>
		/// <param name="name">Name of the setting.</param>
		/// <param name="defaultValue">The default value, if the setting is not found.</param>
		public bool ReadSetting(string section, string name, bool defaultValue)
		{
			// try to read in as string first
			var value = iniFile.ReadString(section, name);

			// check for an empty string -> setting not found
			if (value.Length == 0)
				return defaultValue;

			// first try to parse as a boolean
			bool result;
			if (bool.TryParse(value, out result))
				return result;

			// otherwise parse as int (and fail if not an int)
			return int.Parse(value, iniFile.Culture) != 0 ? true : false;
		}

		/// <summary>Reads in multiple settings from a section in the config file as properties of some class instance.</summary>
		/// <param name="instance">Class instance to read the settings in for.</param>
		/// <param name="section">Section in the config file.</param>
		/// <param name="settings">One or more setting names which must be identical to properties in the class instance.</param>
		public void ReadSettings(object instance, string section, params string[] settings)
		{
			foreach (var setting in settings)
			{
				// find property
				var property = instance.GetType().GetProperty(setting);
				if (property == null)
				{
					Console.WriteLine(string.Format("Warning: Property \"{0}\" not found in {1}.", setting, instance));
					continue;
				}

				// read in setting from config file, using current value of property as default value if it is not found in the config file
				if      (property.PropertyType == typeof(float[])) property.SetValue(instance, ReadSetting(section, setting,  (float[])property.GetValue(instance, null)), null);
				else if (property.PropertyType == typeof(float))   property.SetValue(instance, ReadSetting(section, setting,    (float)property.GetValue(instance, null)), null);
				else if (property.PropertyType == typeof(int))     property.SetValue(instance, ReadSetting(section, setting,      (int)property.GetValue(instance, null)), null);
				else if (property.PropertyType.IsEnum)             property.SetValue(instance, ReadSetting(section, setting,     (Enum)property.GetValue(instance, null)), null);
				else if (property.PropertyType == typeof(bool))    property.SetValue(instance, ReadSetting(section, setting,     (bool)property.GetValue(instance, null)), null);
				else if (property.PropertyType == typeof(Vector2)) property.SetValue(instance, ReadSetting(section, setting, ((Vector2)property.GetValue(instance, null)).ToFloatArray()).ToVector2(), null);
				else if (property.PropertyType == typeof(Vector3)) property.SetValue(instance, ReadSetting(section, setting, ((Vector3)property.GetValue(instance, null)).ToFloatArray()).ToVector3(), null);
				else if (property.PropertyType == typeof(Vector4)) property.SetValue(instance, ReadSetting(section, setting, ((Vector4)property.GetValue(instance, null)).ToFloatArray()).ToVector4(), null);
				else Console.WriteLine(string.Format("Warning: Setting \"{0}\" of type \"{1}\" is not supported.", setting, property.PropertyType));
			}
		}

		#region ReadSetting in default section "Settings"

		/// <summary>Reads a setting as a float array from the config file (section "Setting").</summary>
		/// <param name="name">Name of the setting.</param>
		/// <param name="defaultValue">The default value, if the setting is not found.</param>
		public float[] ReadSetting(string name, float[] defaultValue)
		{
			return ReadSetting("Settings", name, defaultValue);
		}

		/// <summary>Reads a setting as a float from the config file (section "Setting").</summary>
		/// <param name="name">Name of the setting.</param>
		/// <param name="defaultValue">The default value, if the setting is not found.</param>
		public float ReadSetting(string name, float defaultValue)
		{
			return ReadSetting("Settings", name, defaultValue);
		}

		/// <summary>Reads a setting as a int from the config file (section "Setting").</summary>
		/// <param name="name">Name of the setting.</param>
		/// <param name="defaultValue">The default value, if the setting is not found.</param>
		public int ReadSetting(string name, int defaultValue)
		{
			return ReadSetting("Settings", name, defaultValue);
		}

		#endregion
	}
}
