﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: MeshRenderer.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Drawing;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib;
using VideoLib.Frames;
using VideoLib.OpenGL;
using VideoLibPixelFormat = VideoLib.Frames.PixelFormat;

namespace RGBZcamera.Library
{
	public partial class MeshRenderer : Node, IOpenGLNode
	{
		/// <summary>The target size of the output frame.</summary>
		public Size Size { get; set; }

		/// <summary>Location of the camera.</summary>
		public Vector3 Eye { get; set; }

		/// <summary>Point in the centre of the view.</summary>
		public Vector3 LookAt { get; set; }

		/// <summary>Initialises a new instance of the <see cref="MeshRenderer"/> class.</summary>
		public MeshRenderer()
		{
			Size   = new Size(400, 300);
			Eye    = Vector3.Zero;
			LookAt = Vector3.UnitZ;

			AddInputPins("in");
			AddOutputPins("out");
		}

		/// <summary>Initialises the OpenGL state for this node.</summary>
		public override void Init()
		{
			GL.GenBuffers(1, out VertexBufferID);
			GL.GenBuffers(1, out IndexBufferID);

			SetupDepthShaders();
		}

		public override void Process()
		{
			// get latest frames
			var frame = GetInputFrame("in");
			if (PassOnEndOfStreamFrame(frame)) return;
			var points = CastFrameTo<TextureFrame>(frame);

			// update mesh
			MeshTools.UpdateBaseMeshVertexBuffer(points.Width, points.Height, VertexBufferID, out NumVertices);
			MeshTools.UpdateBaseMeshIndexBuffer(points.Width, points.Height, IndexBufferID, out NumIndices);

			SetupFramebufferObject();
			SetupMatrices();
			var result = DrawMesh(points);
			PushFrame(result);
			result.RefDec();
			points.RefDec();
		}

		#region Framebuffer Object (off-screen rendering)
//=================================================================================================

		private int FramebufferID;
		private int ColourRenderbufferID;
		private int DepthRenderbufferID;
		private Size FramebufferSize;

		private void SetupFramebufferObject()
		{
			// correct size already? nothing to do then ...
			if (Size == FramebufferSize) return;

			// delete existing buffers
			if (ColourRenderbufferID != 0) GL.DeleteRenderbuffers(1, ref ColourRenderbufferID);
			if (DepthRenderbufferID  != 0) GL.DeleteRenderbuffers(1, ref DepthRenderbufferID);
			if (FramebufferID        != 0) GL.DeleteFramebuffers( 1, ref FramebufferID);

			// generate a new depth renderbuffer
			GL.GenRenderbuffers(1, out DepthRenderbufferID);
			GL.BindRenderbuffer(RenderbufferTarget.Renderbuffer, DepthRenderbufferID);
			GL.RenderbufferStorage(RenderbufferTarget.Renderbuffer, RenderbufferStorage.DepthComponent24, Size.Width, Size.Height);
			GL.BindRenderbuffer(RenderbufferTarget.Renderbuffer, 0);

			// generate a new framebuffer object
			GL.GenFramebuffers(1, out FramebufferID);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, FramebufferID);
			GL.FramebufferRenderbuffer(FramebufferTarget.Framebuffer, FramebufferAttachment.DepthAttachment, RenderbufferTarget.Renderbuffer, DepthRenderbufferID);

			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			GL.ReadBuffer(ReadBufferMode.ColorAttachment0);

			//Utils.CheckFboStatus(); // not complete, colour buffer is missing
			FramebufferSize = Size;
		}
		
//=================================================================================================
		#endregion

		#region Vertex & Fragment shaders
//=================================================================================================

		private int MeshProgramID;

		private void SetupDepthShaders()
		{
			// load and compile shaders
			int MeshVertexShader    = Utils.CreateShaderFromResource("RGBZcamera.Library.Shaders.Mesh.vert");
			int PhongVertexShader   = Utils.CreateShaderFromResource("RGBZcamera.Library.Shaders.Phong.vert");
			int PhongFragmentShader = Utils.CreateShaderFromResource("RGBZcamera.Library.Shaders.Phong.frag");

			// link shaders
			MeshProgramID = Utils.LinkShaders(MeshVertexShader, PhongVertexShader, PhongFragmentShader);

			// set uniforms
			GL.UseProgram(MeshProgramID);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "coordinates"), 0); // texture unit 0
			GL.Uniform3(GL.GetUniformLocation(MeshProgramID, "lightDirection"), new Vector3(0.35f, 0.35f, 1));
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "lightWorldCoords"), 0);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "lightAbsoluteCoords"), 0);

			//// very specular
			//GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "ambient"), 0.2f);
			//GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "diffuse"), 0.8f);
			//GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "specular"), 0.2f);
			//GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "shininess"), 0.4f);

			// just diffuse
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "ambient"), 0.4f);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "diffuse"), 0.6f);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "specular"), 0.0f);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "shininess"), 1.0f);

			GL.UseProgram(0);
		}

//=================================================================================================
		#endregion

		#region Mesh drawing methods
//=================================================================================================

		private int IndexBufferID;
		private int VertexBufferID;
		private int NumIndices;
		private int NumVertices;

		private void SetupMatrices()
		{
			// generic perspective projection matrix
			var projMat = Matrix4.CreatePerspectiveFieldOfView(MathHelper.PiOver4, (float)Size.Width / Size.Height, 0.1f, 10);

			// Y flip so that ReadPixels returns the image the right way up
			projMat.M12 *= -1;
			projMat.M22 *= -1;
			projMat.M32 *= -1;
			projMat.M42 *= -1;

			// simple lookat modelview matrix
			var viewMat = Matrix4.LookAt(Eye, LookAt, Vector3.UnitY);

			// load matrices
			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadMatrix(ref projMat);
			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadMatrix(ref viewMat);
		}

		private TextureFrame DrawMesh(TextureFrame points)
		{
			// allocate the output texture
			var result = new TextureFrame(Size.Width, Size.Height) { InternalFormat = PixelInternalFormat.Rgba16f, SequenceNumber = points.SequenceNumber };
			result.Alloc();

			// set up FBO
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, FramebufferID);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, result.Texture, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();

			// draw depth mesh
			GL.Viewport(0, 0, Size.Width, Size.Height);
			GL.ClearColor(Color.Black);
			GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
			GL.Enable(EnableCap.DepthTest);

			// push the current Array Buffer state so we can restore it later
			GL.PushClientAttrib(ClientAttribMask.ClientVertexArrayBit);

			// sets the vertex pointer to the currently bound array describing how the data is stored
			GL.BindBuffer(BufferTarget.ArrayBuffer, VertexBufferID);
			GL.VertexPointer(3, VertexPointerType.Float, Vector3.SizeInBytes, IntPtr.Zero);
			GL.EnableClientState(ArrayCap.VertexArray);

			GL.UseProgram(MeshProgramID);
			points.Bind(TextureUnit.Texture0);
			//GL.PolygonMode(MaterialFace.FrontAndBack, PolygonMode.Line); // wireframe

			// draw the triangles in the element array buffer
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, IndexBufferID);
			GL.DrawElements(BeginMode.Triangles, NumIndices, DrawElementsType.UnsignedInt, IntPtr.Zero);

			// restore the state
			GL.UseProgram(0);
			points.Unbind();
			GL.PopClientAttrib();
			//GL.PolygonMode(MaterialFace.FrontAndBack, PolygonMode.Fill);

			return result;
		}

//=================================================================================================
		#endregion
	}
}
