﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: BackgroundSubtraction.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

// input textures
uniform sampler2D colour;
uniform sampler2D points;

uniform vec4 backgroundPlane;
uniform float planeThreshold;
uniform bool showMatte;
uniform vec3 backgroundColour;
uniform int mattingTechnique;

// forward declaration of mesh vertex shader functions (from Mesh.vert)
vec3 getVertex(sampler2D depthmap, vec2 position);

void main (void)
{
	// look up texture size once and reuse it later, assuming every texture has the same size
	vec2 size = vec2(textureSize2D(colour, 0));
	
	// fetch input pixel colour
	vec3 shaded = texture2D(colour, gl_FragCoord.xy / size).rgb;
	
	// compute normal
	vec2 pos = gl_FragCoord.xy - 1; // minus one to distribute border evenly on all sides
	vec3 vertex = getVertex(points, pos);
	
	// normalise plane normal (so that the distance to the plane is meaningful)
	vec4 plane = backgroundPlane / length(backgroundPlane.xyz);

	// matte: 0 is background, 1 is foreground
	float matte = 0.0;
	
	switch(mattingTechnique)
	{
		case 0:	// hard binary matte
			matte = dot(vec4(vertex, 1.0), backgroundPlane) < planeThreshold ? 0.0 : 1.0;
			break;
		
		case 1: // blurred binary matte
			matte += dot(vec4(getVertex(points, pos + vec2(-1,-1)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 1.0 / 16.0;
			matte += dot(vec4(getVertex(points, pos + vec2(-1, 0)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 2.0 / 16.0;
			matte += dot(vec4(getVertex(points, pos + vec2(-1, 1)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 1.0 / 16.0;
			matte += dot(vec4(getVertex(points, pos + vec2( 0,-1)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 2.0 / 16.0;
			matte += dot(vec4(getVertex(points, pos + vec2( 0, 0)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 4.0 / 16.0;
			matte += dot(vec4(getVertex(points, pos + vec2( 0, 1)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 2.0 / 16.0;
			matte += dot(vec4(getVertex(points, pos + vec2( 1,-1)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 1.0 / 16.0;
			matte += dot(vec4(getVertex(points, pos + vec2( 1, 0)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 2.0 / 16.0;
			matte += dot(vec4(getVertex(points, pos + vec2( 1, 1)), 1.0), backgroundPlane) < planeThreshold ? 0.0 : 1.0 / 16.0;
			break;
	
		case 2: // simple soft matte
			matte = dot(vec4(vertex, 1.0), backgroundPlane) / planeThreshold - 0.5;
			break;
	}
	
	matte = clamp(matte, 0.0, 1.0);
	
	// output matte
	if(showMatte) gl_FragColor = vec4(vec3(matte), 1.0);
	else          gl_FragColor = vec4(mix(backgroundColour, shaded, vec3(matte)), 1.0);
}
