﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: SpatioTemporalFilter.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

// input textures
uniform sampler2D currentVideo;
uniform sampler2D previousVideo;
uniform sampler2D currentDistance;
uniform sampler2D previousDistance;
uniform sampler2D opticalFlow;

// filter parameters
uniform float colourSigma;
uniform float distanceSigma;
uniform float flowSigma;
uniform float spatialSigma;
uniform float falloffWeight;

void main()
{
	// look up texture size once and reuse it later,
	// assuming every texture has the same size
	vec2 size = textureSize2D(currentVideo, 0);
	
	// precompute 1 / sigma^2
	float iSigmaS = 1.0 / (2 * spatialSigma  * spatialSigma);
	float iSigmaF = 1.0 / (2 * flowSigma     * flowSigma);
	float iSigmaC = 1.0 / (2 * colourSigma   * colourSigma);
	float iSigmaD = 1.0 / (2 * distanceSigma * distanceSigma);

	// truncate Gaussian after 2 sigma
	int radius = int(ceil(2 * spatialSigma));
	
	// homogeneous accumulator of values, i.e. (w * value, w)
	vec2 acc = vec2(0.0);
	
	// colour and distance values at the central pixel
	vec3 centreC = texture2D(currentVideo,    gl_FragCoord.xy / size).rgb;
	vec4 centreD = texture2D(currentDistance, gl_FragCoord.xy / size);
	
	// motion-compensated centre of kernel in previous frame
	vec4 centreFlow = texture2D(opticalFlow, gl_FragCoord.xy / size);
	vec2 previousCentreCoords = vec2(gl_FragCoord.x - centreFlow.x, gl_FragCoord.y + centreFlow.y) / size;

	// loop over rectangular 2D kernel of pixels centred around gl_FragCoord.xy
	for(float yi = gl_FragCoord.y - radius; yi <= gl_FragCoord.y + radius; yi++)
	{
		for(float xi = gl_FragCoord.x - radius; xi <= gl_FragCoord.x + radius; xi++)
		{
			// ignore out of range pixels
			if(xi >= 0 && xi < size.x && yi >= 0 && yi < size.y)
			{
				//---- current frame --------------------------------------------------------------

				// distance of kernel pixel in current frame
				vec2 currentCoords = vec2(xi, yi) / size;
				float currentD = texture2D(currentDistance, currentCoords).w;
				
				// optical flow at the kernel pixel
				vec4 flows = texture2D(opticalFlow, currentCoords);
				vec2 previousCoords = vec2(xi - flows.x, yi + flows.y) / size;

				// ignore pixels with invalid depth
				if(currentD != 0.0f)
				{
					// colour of kernel pixel in current frame
					vec3 currentC = texture2D(currentVideo, currentCoords).rgb;
					
					// compute pixel weight, using exp(a) * exp(b) = exp(a + b);
					float w = 0.0f;

					// spatial weight
					vec2 deltaS = vec2(xi, yi) - gl_FragCoord.xy;
					w += iSigmaS * dot(deltaS, deltaS);

					// colour weight
					vec3 deltaC = currentC - centreC;
					w += clamp(2 - length(flows.zw) / flowSigma, 0.0, 1.0) * iSigmaC * dot(deltaC, deltaC);

					// distance weight (ignore if central pixel invalid)
					float deltaD = currentD - centreD.w;
					if(centreD.w != 0.0f) w += iSigmaD * (deltaD * deltaD);

					acc += exp(-w) * (1 - falloffWeight) * vec2(currentD, 1);
				}
				
				//---- previous frame -------------------------------------------------------------
				
				//continue;
				if(falloffWeight == 0.0) continue;
								
				if(previousCoords.x < 0 || previousCoords.x >= 1) continue;
				if(previousCoords.y < 0 || previousCoords.y >= 1) continue;
				
				// distance of kernel pixel in current frame
				float previousD = texture2D(previousDistance, previousCoords).w;
				
				// ignore pixels with invalid depth
				if(previousD != 0.0f)
				{
					// colour of kernel pixel in previous frame
					vec3 previousC = texture2D(previousVideo, previousCoords).rgb;
					
					// compute pixel weight, using exp(a) * exp(b) = exp(a + b);
					float w = 0.0f;

					// spatial weight
					//vec2 deltaS = vec2(xi, yi) - gl_FragCoord.xy;
					vec2 deltaS = size * (previousCoords - previousCentreCoords);
					w += iSigmaS * dot(deltaS, deltaS);
					
					// flow weight
					w += iSigmaF * dot(flows.xy, flows.xy);
					
					// colour weight
					vec3 deltaC = previousC - centreC;
					w += iSigmaC * dot(deltaC, deltaC);

					// distance weight (ignore if central pixel invalid)
					float deltaD = previousD - centreD.w;
					if(centreD.w != 0.0f) w += iSigmaD * (deltaD * deltaD);
					
					acc += exp(-w) * falloffWeight * vec2(previousD, 1);
				}
			}
		}
	}
	
	// write back filtered point
	if(acc.y <= 0.0)
	{
		// just return black if there were no valid data samples in the kernel
		gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);
	}
	else
	{
		// distance of the old point (need to compute it in case of invalid point)
		float old_dist = (centreD.w == 0.0 ? sqrt(dot(centreD, centreD)) : centreD.w);

		//// Exponential History Buffer and Temporal Weighting Function
		//// (per Herzog et al.'s "Spatio-Temporal Upsampling on the GPU")
		//
		//float htilde = (acc.x / acc.y);
		//float wtilde = acc.y;
		//
		//vec4 flows = texture2D(opticalFlow, gl_FragCoord.xy / size);
		//vec2 previousCoords = vec2(gl_FragCoord.x - flows.x, gl_FragCoord.y + flows.y) / size;
		//vec4 history = texture2D(previousDistance, previousCoords);
		//float h_r = length(history.xyz);
		//float h_w = history.w;
		//
		//// estimating temporal gradients
		//float a = falloffWeight;
		//float oneMinusAbsGradient = 1 - abs((htilde - h_r) / max(abs(htilde), abs(h_r)));
		//float w_f = a * oneMinusAbsGradient * oneMinusAbsGradient;
		//// w_f = falloffWeight;
		//
		//float h = (htilde * wtilde + w_f * h_w * h_r) / (wtilde + w_f * h_w);

		// new, filtered distance
		float new_dist = (acc.x / acc.y);
		//new_dist = h;

		// rescale input point and set new distance
		vec4 new_coords = centreD / old_dist * new_dist;
		new_coords.w = new_dist;
		//new_coords.w = wtilde + falloffWeight * h_w;

		// write the point
		gl_FragColor = new_coords;
	}
		
	//---- FlowAverager ---------------------------------------------------------------------------
		
	//vec4 flows = texture2D(opticalFlow, gl_FragCoord.xy / textureSize2D(opticalFlow, 0));
	//vec2 coords = gl_FragCoord.xy - vec2(1, -1) * flows.xy;
	//
	//vec4 old = texture2D(previousDistance, coords          / textureSize2D(previousDistance, 0));
	//vec4 new = texture2D(currentDistance,  gl_FragCoord.xy / textureSize2D(currentDistance,  0));
	//
	////float w = 1; // no filtering (baseline)
	//float w = weight; // constant weight
	////float w = mix(weight, 1, clamp(10 * abs(old.w - new.w), 0, 1)); // adaptive weight
	//
	//gl_FragColor = mix(old, new, w);
}

