﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: SpatioTemporalFilter.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.OpenGL;
using OpenTK.Graphics.OpenGL;
using System.ComponentModel;

namespace RGBZcamera.Library
{
	/// <summary>Implements a spatio-temporal bilateral filter for upsampling and denoising depth video data.</summary>
	public class SpatioTemporalFilter : FboNode
	{
		/// <summary>Gets or sets the spatial sigma which prefers closeby pixels.</summary>
		[Uniform(), DefaultValue(5.0f), Description("The spatial sigma prefers closeby pixels.")]
		public float SpatialSigma { get; set; }

		/// <summary>Gets or sets the flow sigma which prefers history samples with small flow between frames.</summary>
		/// <value>The flow sigma.</value>
		[Uniform(), DefaultValue(5.0f), Description("The flow sigma prefers history samples with small flow between frames.")]
		public float FlowSigma { get; set; }

		/// <summary>Gets or sets the colour sigma which prefers pixels similar in colour.</summary>
		[Uniform(), DefaultValue(0.1f), Description("The colour sigma prefers pixels similar in colour.")]
		public float ColourSigma { get; set; }

		/// <summary>Gets or sets the distance sigma which prefers pixels at a similar distance.</summary>
		[Uniform(), DefaultValue(0.1f), Description("The distance sigma prefers pixels at a similar distance.")]
		public float DistanceSigma { get; set; }

		/// <summary>Gets or sets the falloff weight which prefers smoothed history data.</summary>
		[Uniform(), DefaultValue(0.9f), Description("The fall-off weight prefers smoothed history data.")]
		public float FalloffWeight { get; set; }

		/// <summary>Gets or sets a value indicating whether this <see cref="SpatioTemporalFilter"/> node is enabled.</summary>
		[DefaultValue(true)]
		public bool Enabled { get; set; }

		/// <summary>Initialises a new instance of the <see cref="SpatioTemporalFilter"/> class.</summary>
		public SpatioTemporalFilter()
		{
			// default settings
			SpatialSigma  = 5;
			FlowSigma     = 5;
			ColourSigma   = 0.1f;
			DistanceSigma = 0.1f;
			FalloffWeight = 0.9f;
			Enabled       = true;

			AddInputPins("currentVideo", "previousVideo", "distance", "flow");
			AddOutputPin("out");
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.SpatioTemporalFilter.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Name = "currentVideo",  Unit = 0 },
				new TextureBinding { Name = "previousVideo", Unit = 1 },
				new TextureBinding { Pin = "flow",     Uniform = "opticalFlow",      Unit = 2 },
				new TextureBinding { Pin = "distance", Uniform = "currentDistance",  Unit = 3 },
				new TextureBinding { Pin = null,       Uniform = "previousDistance", Unit = 4 },
			};

			base.Init();
		}

		private TextureFrame previousAverage;

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get latest frame
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);
			AssertFrameProperties(textures, FrameProps.Size);

			// draw quads
			TextureFrame result;
			if (Enabled)
			{
				result = Draw(textures);
			}
			else
			{
				result = textures["distance"];
				result.RefInc();
			}
			PushFrame(result);

			// free previous frame
			if (previousAverage != null)
				previousAverage.RefDec();

			// keep the result for the next frame
			previousAverage = result;

			// clean up
			textures.RefDec();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			// allocate a texture that is identical to the input
			var resultTexture = TextureFrame.CreateToMatch(inputs["distance"]);
			resultTexture.Wrap = TextureWrapMode.ClampToEdge;
			resultTexture.Alloc();

			// bind textures
			if (previousAverage != null) previousAverage.Bind(TextureUnit.Texture4);

			// for first frame, use weight = 1 to ignore previous average (as it is undefined)
			float w = 0;
			if (previousAverage == null) { w = FalloffWeight; FalloffWeight = 0; }

			// bind textures, update uniforms and draw to FBO
			Draw(inputs, resultTexture);

			// return to actual weight for next frame
			if (previousAverage == null) FalloffWeight = w;

			return resultTexture;
		}
	}
}
