﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: WinnemoellerAbstraction.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib;
using VideoLib.OpenGL;
using System.ComponentModel;
using OpenTK.Graphics.OpenGL;

namespace RGBZcamera.Library.WinnemoellerAbstraction
{
	/// <summary>Implement Real-Time Video Abstraction (Winnemöller et al. 2006).</summary>
	public class WinnemoellerAbstraction : Node, IOpenGLNode
	{
		[Category("1. Abstraction"), DefaultValue(4.25f), DisplayName("sigma_r"), Description("Range sigma of the bilateral abstraction step. Paper: 4.25.")]
		public float SigmaR { get; set; }

		[Category("1. Abstraction"), DefaultValue(3.0f), DisplayName("sigma_d"), Description("Distance sigma of the bilateral abstraction step. Paper: 3.")]
		public float SigmaD { get; set; }

		[Category("1. Abstraction"), DefaultValue(2), DisplayName("n_e"), Description("Number of bilateral filter iterations for edge extraction. Paper: 1 or 2.")]
		public int N_E { get; set; } // edge iterations

		[Category("1. Abstraction"), DefaultValue(4), DisplayName("n_b"), Description("Number of bilateral filter iterations for smoothed image. Paper: 3 or 4.")]
		public int N_B { get; set; } // smoothing iterations

		[Category("1. Abstraction"), DefaultValue(1.0f), Description("Modifies the saturation of the video prior to processing.")]
		public float Saturation { get; set; }

		[Category("3. Soft quantisation"), DefaultValue(8), DisplayName("q"), Description("Number of quantisation steps. Paper: [8,10].")]
		public int Q { get; set; }

		[Category("3. Soft quantisation"), DefaultValue(0.0f), DisplayName("Lambda_delta"), Description("Refer to paper. Useful range: 0..5.")]
		public float LambdaDelta { get; set; }

		[Category("3. Soft quantisation"), DefaultValue(2.0f), DisplayName("Omega_delta"), Description("Refer to paper. Useful range: 0..5.")]
		public float OmegaDelta { get; set; }

		[Category("3. Soft quantisation"), DefaultValue(0.9f), DisplayName("Lambda_phi"), Description("Refer to paper. Useful range: 0..1000.")]
		public float LambdaPhi { get; set; }

		[Category("3. Soft quantisation"), DefaultValue(1.6f), DisplayName("Omega_phi"), Description("Refer to paper. Useful range: 0..1000.")]
		public float OmegaPhi { get; set; }

		[Category("2. DoG edges"), DefaultValue(2.0f), DisplayName("sigma_e"), Description("Edge sigma for DoG edges. Paper: 2 (fine), 5 (coarse).")]
		public float SigmaE { get; set; }

		[Category("2. DoG edges"), DefaultValue(1.0f), DisplayName("phi_e"), Description("DoG edge falloff. Paper: [0.75,5].")]
		public float PhiE { get; set; }

		[Category("2. DoG edges"), DefaultValue(0.98f), DisplayName("tau"), Description("Refer to paper. Useful range: 0..1.")]
		public float Tau { get; set; }

		private RGB2Lab                  rgb2lab;
		private Lab2RGB                  lab2rgb;
		private SeparableBilateralX      bilateralX;
		private SeparableBilateralY      bilateralY;
		private GaussianX                gaussianX;
		private GaussianY                gaussianY;
		private WinnemoellerDogEdges     edges;
		private WinnemoellerQuantisation quant;
		private WinnemoellerCombine      combine;

		/// <summary>Initialises a new instance of the <see cref="WinnemoellerAbstraction"/> class.</summary>
		/// <remarks>To link a node to a different topology, just assign to Node.Topology.</remarks>
		public WinnemoellerAbstraction()
		{
			// default parameters
			SigmaR      = 4.25f;
			SigmaD      = 3;
			N_E         = 2;
			N_B         = 4;
			Saturation  = 1;
			SigmaE      = 2.0f;
			Q           = 8;
			LambdaDelta = 0;
			OmegaDelta  = 2;
			LambdaPhi   = 0.9f;
			OmegaPhi    = 1.6f;
			PhiE        = 1.0f;
			Tau         = 0.98f;

			// create nodes outside the current topology (parameters are set later)
			rgb2lab    = new RGB2Lab                  { Topology = null };
			lab2rgb    = new Lab2RGB                  { Topology = null };
			bilateralX = new SeparableBilateralX      { Topology = null };
			bilateralY = new SeparableBilateralY      { Topology = null };
			gaussianX  = new GaussianX                { Topology = null };
			gaussianY  = new GaussianY                { Topology = null };
			edges      = new WinnemoellerDogEdges     { Topology = null };
			quant      = new WinnemoellerQuantisation { Topology = null };
			combine    = new WinnemoellerCombine      { Topology = null };

			AddDefaultPins();
		}

		/// <summary>Initialisation code, executed just before the first StatusUpdate.</summary>
		public override void Init()
		{
			rgb2lab.Init();
			lab2rgb.Init();
			bilateralX.Init();
			bilateralY.Init();
			gaussianX.Init();
			gaussianY.Init();
			edges.Init();
			quant.Init();
			combine.Init();
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get input
			var frame = GetInputFrame();
			if (PassOnEndOfStreamFrame(frame)) return;
			var texture = CastFrameTo<TextureFrame>(frame);

			// convert to CIELAB
			rgb2lab.Saturation = Saturation;
			var tmpTex1 = rgb2lab.Draw(new Dictionary<string, TextureFrame> { { "in", texture } });
			texture.RefDec();

			// apply bilateral filter iteratively
			TextureFrame smoothInput = null, edgeInput = null;
			bilateralX.SigmaD = SigmaD; bilateralX.SigmaR = SigmaR;
			bilateralY.SigmaD = SigmaD; bilateralY.SigmaR = SigmaR;

			// NB: paper says ne < nb, but this works even if not
			if (N_B == 0) { smoothInput = tmpTex1; smoothInput.RefInc(); }
			if (N_E == 0) { edgeInput   = tmpTex1; edgeInput.RefInc();   }
			for (int i = 0; i < Math.Max(N_E, N_B); i++)
			{
				// first filter horizontally
				var tmpHorz = bilateralX.Draw(new Dictionary<string, TextureFrame> { { "in", tmpTex1 } });
				tmpTex1.RefDec();

				// then vertically
				tmpTex1 = bilateralY.Draw(new Dictionary<string, TextureFrame> { { "in", tmpHorz } });
				tmpHorz.RefDec();

				if (i == N_B - 1) { smoothInput = tmpTex1; smoothInput.RefInc(); }
				if (i == N_E - 1) { edgeInput   = tmpTex1; edgeInput.RefInc();   }
			}
			tmpTex1.RefDec();

			// compute DoG edges
			gaussianX.Sigma = SigmaE;
			gaussianY.Sigma = SigmaE;
			tmpTex1         = gaussianX.Draw(new Dictionary<string, TextureFrame> { { "in", edgeInput } });
			var gaussCentre = gaussianY.Draw(new Dictionary<string, TextureFrame> { { "in", tmpTex1 } });
			tmpTex1.RefDec();

			// surround (sigma = sqrt(1.6) * sigma_e)
			gaussianX.Sigma   = (float)Math.Sqrt(1.6) * SigmaE;
			gaussianY.Sigma   = (float)Math.Sqrt(1.6) * SigmaE;
			tmpTex1           = gaussianX.Draw(new Dictionary<string, TextureFrame> { { "in", edgeInput } });
			var gaussSurround = gaussianY.Draw(new Dictionary<string, TextureFrame> { { "in", tmpTex1 } });
			edgeInput.RefDec();
			tmpTex1.RefDec();

			// DoG edges
			edges.PhiE = PhiE;
			edges.Tau  = Tau;
			tmpTex1 = edges.Draw(new Dictionary<string, TextureFrame> { { "inCentre", gaussCentre }, { "inSurround", gaussSurround } });
			gaussSurround.RefDec();
			gaussCentre.RefDec();

			// apply luminance quantisation
			quant.Q           = Q;
			quant.LambdaDelta = LambdaDelta;
			quant.OmegaDelta  = OmegaDelta;
			quant.LambdaPhi   = LambdaPhi;
			quant.OmegaPhi    = OmegaPhi;
			var quantised = quant.Draw(new Dictionary<string, TextureFrame> { { "in", smoothInput } });
			smoothInput.RefDec();

			// combine smoothed image and edges
			var tmpTex2 = combine.Draw(new Dictionary<string, TextureFrame> { { "inSmooth", quantised }, { "inEdges", tmpTex1 } });
			quantised.RefDec();
			tmpTex1.RefDec();

			//-- convert back to sRGB
			var result = lab2rgb.Draw(new Dictionary<string, TextureFrame> { { "in", tmpTex2 } });
			tmpTex2.RefDec();

			PushFrame(result);
			result.RefDec();
		}
	}
}
