﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: RecorderProgram.cs 549 2012-04-16 12:56:15Z cr333 $
*/

// Comment this line out if only Kinect support is needed.
//#define CompileFleaSR4000

using System;
using System.IO;
using System.Security.Permissions;
using System.Threading;
using System.Windows.Forms;
using VideoLib;
using VideoLib.Filters;
using VideoLib.FlyCapture;
using VideoLib.Frames;
using VideoLib.LibAv;
using VideoLib.OpenNI;
using VideoLib.Scheduling;
using VideoLib.Sinks;
using VideoLib.Sources;
using VideoLib.Utils;

#if CompileFleaSR4000
using VideoLib.SR4000;
#endif

namespace RGBZcamera.Recorder
{
	class RecorderProgram
	{
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main()
		{
			Application.EnableVisualStyles();
			Application.SetCompatibleTextRenderingDefault(false);
			Application.Run(new RecorderForm());
		}

		private static RecorderForm dataDisplay;
		private static Scheduler scheduler;
		private static Topology topology;

		public static void Start(RecorderForm form, Settings settings, bool record)
		{
			// set up topology and scheduler ------------------------------------------------------
			dataDisplay = form;
			topology = Topology.MainInstance;
			scheduler = new SimpleDequeScheduler(topology) { Profiling = true };

			// convert raw to images (for calibration) --------------------------------------------

			//string path = @"C:\Users\cr333\Documents\Projects\RGBZCamera\data\input\tableB\";

			//var amplitude = topology.Connect(new RawSource(path + @"amplitude-{0:D4}.raw"), new ConvertFilter(PixelFormat.Grey8) { ScaleFactor = 10 });
			//var depth = topology.Connect(new RawSource(path + @"depth-{0:D4}.raw"), new ConvertFilter(PixelFormat.Grey8) { ScaleFactor = 5 }); // * 6.5536f  for kinect data
			//var video = topology.Connect(new RawSource(path + @"video-{0:D4}.raw"), new LibAvConvert(PixelFormat.BGR24));

			//topology.ConnectFan(amplitude, new ImageSequenceSink(path + @"images\amplitude-{0:D4}.png"), new VideoDisplay(form.videoPanel2, "Amplitude"));
			//topology.ConnectFan(depth, new ImageSequenceSink(path + @"images\depth-{0:D4}.png"), new VideoDisplay(form.videoPanel1, "Depth"));
			//topology.ConnectFan(video, new ImageSequenceSink(path + @"images\video-{0:D4}.png"), new VideoDisplay(form.videoPanel4, "Video"));

			// set up nodes -----------------------------------------------------------------------

			#if CompileFleaSR4000
				SwissRanger4000Source sr4000 = null;
				FlyCaptureSource flea2       = null;
			#endif

			KinectSource kinect          = null;

			switch (settings.CaptureSource)
			{
				#if CompileFleaSR4000
					case VideoSource.FleaSR4000:

						sr4000 = new SwissRanger4000Source()
						{
							AutoExposure    = settings.AutoExposureSR,
							IntegrationTime = settings.IntegrationTime,
							Coordinates     = settings.CaptureCoords,
							ConfidenceMap   = settings.CaptureConfidence,
							Triggered = true
						};

						flea2 = new FlyCaptureSource()
						{
							Width            = settings.Width,
							CaptureFormat    = settings.Format,
							Framerate        = settings.Framerate,
							AutoExposure     = settings.AutoExposureFlea,
							Exposure         = settings.Exposure,
							AutoGain         = settings.AutoGain,
							Gain             = settings.Gain,
							Gamma            = settings.Gamma,
							AutoSaturation   = settings.AutoSaturation,
							Saturation       = settings.Saturation,
							AutoShutter      = settings.AutoShutter,
							Shutter          = settings.Shutter,
							AutoWhiteBalance = settings.AutoWhiteBalance,
							WhiteBalanceBlue = settings.WhiteBalanceBlue,
							WhiteBalanceRed  = settings.WhiteBalanceRed,
							Strobe = true
						};

						break;
				#endif // CompileFleaSR2000

				case VideoSource.Kinect:

					kinect = new KinectSource(settings.Streams)
					{
						ImageMode = settings.ImageMode,
						IRMode    = settings.IRMode,
						DepthMode = settings.DepthMode,
						SyncMode  = settings.SyncMode,
					};

					break;

				default:
					throw new ArgumentException("Unknown video source.");
			}

			if (settings.CaptureDisplay)
			{
				switch (settings.CaptureSource)
				{
					#if CompileFleaSR4000
						case VideoSource.FleaSR4000:
							topology.Connect(sr4000["depth"], new VideoDisplay(form.videoPanel1, "Depth"));
							topology.Connect(sr4000["amplitude"], new ConvertFilter() { ScaleFactor = 6 }, new VideoDisplay(form.videoPanel2, "Amplitude"));
							topology.Connect(sr4000["confidence"], new VideoDisplay(form.videoPanel3, "Confidence"));
							topology.Connect(flea2, new VideoDisplay(form.videoPanel4, "Video"));
						break;
					#endif

					case VideoSource.Kinect:

						// if image is streamed, it is always shown big
						if ((settings.Streams & Streams.Image) == Streams.Image)
							topology.Connect(kinect["image"], new VideoDisplay(form.videoPanel4, "Video"));

						// if IR is streamed, it is always shown big (mutually exclusive with image)
						if ((settings.Streams & Streams.IR) == Streams.IR)
							topology.Connect(kinect["ir"], new ConvertFilter() { ScaleFactor = 2000, Offset = -0.1f }, new VideoDisplay(form.videoPanel4, "IR"));

						// show depth big if it's the only stream captured, otherwise small
						if ((settings.Streams & Streams.Depth) == Streams.Depth)
						{
							if(settings.Streams == Streams.Depth) // only depth stream? -> show big
								topology.Connect(kinect["depth"], new ConvertFilter() { ScaleFactor = 6.5536f }, new VideoDisplay(form.videoPanel4, "Depth"));
							else // show depth as small video
								topology.Connect(kinect["depth"], new ConvertFilter() { ScaleFactor = 6.5536f }, new VideoDisplay(form.videoPanel2, "Depth"));
						}
						break;

					default:
						break;
				}
				
			}

			if (record)
			{
				String outputDir = Path.Combine(settings.CapturePath, DateTime.Now.ToString("yyyy-MM-dd HH-mm-ss"));
				Directory.CreateDirectory(outputDir);
				settings.SaveToFile(Path.Combine(outputDir, "_settings.txt"));

				string ext = settings.CaptureFormat.ToString().ToLower();
				switch (settings.CaptureSource)
				{
					#if CompileFleaSR4000
						case VideoSource.FleaSR4000:
							if (ext == "raw")
							{
								if (settings.CaptureDepth)      topology.Connect(sr4000["depth"],      new RawSink(outputDir + @"\depth-{0:D4}."      + ext) { Async = settings.CaptureAsync, BufferFrames = settings.CaptureBuffered });
								if (settings.CaptureAmplitude)  topology.Connect(sr4000["amplitude"],  new RawSink(outputDir + @"\amplitude-{0:D4}."  + ext) { Async = settings.CaptureAsync, BufferFrames = settings.CaptureBuffered });
								if (settings.CaptureConfidence) topology.Connect(sr4000["confidence"], new RawSink(outputDir + @"\confidence-{0:D4}." + ext) { Async = settings.CaptureAsync, BufferFrames = settings.CaptureBuffered });
								if (settings.CaptureCoords)     topology.Connect(sr4000["coords"],     new RawSink(outputDir + @"\coords-{0:D4}."     + ext) { Async = settings.CaptureAsync, BufferFrames = settings.CaptureBuffered });
								topology.Connect(flea2, new RawSink(outputDir + @"\video-{0:D4}." + ext) { Async = true, BufferFrames = settings.CaptureBuffered });
							}
							else // save as images frames
							{
								if (settings.CaptureDepth)      topology.Connect(sr4000["depth"],      new ConvertFilter(PixelFormat.Grey8),                     new ImageSequenceSink(outputDir + @"\depth-{0:D4}." + ext));
								if (settings.CaptureAmplitude)  topology.Connect(sr4000["amplitude"],  new ConvertFilter(PixelFormat.Grey8) { ScaleFactor = 2 }, new ImageSequenceSink(outputDir + @"\amplitude-{0:D4}." + ext));
								if (settings.CaptureConfidence) topology.Connect(sr4000["confidence"], new ConvertFilter(PixelFormat.Grey8),                     new ImageSequenceSink(outputDir + @"\confidence-{0:D4}." + ext));
								topology.Connect(flea2, new ImageSequenceSink(outputDir + @"\video-{0:D4}." + ext));
								// no coords saved!
							}
							break;
					#endif // CompileFleaSR4000

					case VideoSource.Kinect:
						if (ext == "raw")
						{
							if ((settings.Streams & Streams.Image) == Streams.Image)
								topology.Connect(kinect["image"], new RawSink(outputDir + @"\video-{0:D4}." + ext) { Async = true, BufferFrames = settings.CaptureBuffered });

							if ((settings.Streams & Streams.IR) == Streams.IR)
								topology.Connect(kinect["ir"], new RawSink(outputDir + @"\ir-{0:D4}." + ext) { Async = settings.CaptureAsync, BufferFrames = settings.CaptureBuffered });

							if ((settings.Streams & Streams.Depth) == Streams.Depth)
								topology.Connect(kinect["depth"], new RawSink(outputDir + @"\depth-{0:D4}." + ext) { Async = settings.CaptureAsync, BufferFrames = settings.CaptureBuffered });
						}
						else // save as images frames
						{
							if ((settings.Streams & Streams.Image) == Streams.Image)
								topology.Connect(kinect["image"], new LibAvConvert(PixelFormat.BGR24), new ImageSequenceSink(outputDir + @"\video-{0:D4}." + ext));

							if ((settings.Streams & Streams.IR) == Streams.IR)
								topology.Connect(kinect["ir"], new ConvertFilter(PixelFormat.Grey8) { ScaleFactor = 2000, Offset = -0.1f }, new ImageSequenceSink(outputDir + @"\ir-{0:D4}." + ext));

							if ((settings.Streams & Streams.Depth) == Streams.Depth)
								topology.Connect(kinect["depth"], new ConvertFilter(PixelFormat.Grey8) { ScaleFactor = 6.5536f }, new ImageSequenceSink(outputDir + @"\depth-{0:D4}." + ext));
						}
						break;

					default:
						break;
				}

			}

			// start the scheduler ----------------------------------------------------------------
			scheduler.Setup();
			scheduler.StartWorkers(4);
			scheduler.Join();

			if (kinect != null)
			{
				kinect.Dispose();
				kinect = null;
			}
		}

		public static void Stop(RecorderForm form)
		{
			scheduler.Stop();
			topology.Clear();
		}


		[SecurityPermission(SecurityAction.Demand, Flags = SecurityPermissionFlag.ControlAppDomain)]
		public static void RegisterUnhandledExceptionEventHandler()
		{
			AppDomain currentDomain = AppDomain.CurrentDomain;
			currentDomain.UnhandledException += new UnhandledExceptionEventHandler(AppDomain_UnhandledException);
		}

		static void Application_ThreadException(object sender, ThreadExceptionEventArgs args)
		{
			Exception e = args.Exception;
			MessageBox.Show(
				String.Format("{0}: {1}\n\n{2}\n\n(Application.ThreadException)", e.GetType().Name, e.Message, e),
				String.Format("Error: {0}", e.GetType()), MessageBoxButtons.OK, MessageBoxIcon.Error);
			dataDisplay.BeginInvoke(new Action(delegate() { dataDisplay.Close(); }));
		}

		static void AppDomain_UnhandledException(object sender, UnhandledExceptionEventArgs args)
		{
			Exception e = (Exception)args.ExceptionObject;
			MessageBox.Show(
				String.Format("{0}: {1}\n\n{2}\n\n(AppDomain.UnhandledException)", e.GetType().Name, e.Message, e),
				String.Format("Error: {0}", e.GetType()), MessageBoxButtons.OK, MessageBoxIcon.Error);
			dataDisplay.BeginInvoke(new Action(delegate() { dataDisplay.Close(); }));
		}
	}
}
