﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: InputVideo.cs 535 2012-04-04 14:50:13Z cr333 $
*/
using System.IO;
using RGBZcamera.Library.Config;

namespace RGBZcamera.Renderer
{
	/// <summary>Wraps the FileInfo object corresponding to a config file and displays as "name (nnn frames)" in a ComboBox.</summary>
	class InputVideo
	{
		public InputVideo(FileInfo configFile)
		{
			Config = new ConfigReader(configFile.FullName);
			Frames = Config.Video.LastFrame - Config.Video.FirstFrame + 1;
		}

		public ConfigReader Config { get; set; }
		public int Frames { get; set; }

		/// <summary>Indicates if the video frames are found in the expected place.</summary>
		public bool DataFound
		{
			get
			{
				// look up cached value to avoid unnecessary disk accesses
				if (dataFound != null)
					return (bool)dataFound;

				dataFound = true;

				// check if the first and last files of the colour and depth videos exist
				if (!File.Exists(Path.Combine(Config.InputDir, string.Format(Config.Video.Files, Config.Video.FirstFrame)))) { dataFound = false; return false; }
				if (!File.Exists(Path.Combine(Config.InputDir, string.Format(Config.Video.Files, Config.Video.LastFrame))))  { dataFound = false; return false; }
				if (!File.Exists(Path.Combine(Config.InputDir, string.Format(Config.Depth.Files, Config.Depth.FirstFrame)))) { dataFound = false; return false; }
				if (!File.Exists(Path.Combine(Config.InputDir, string.Format(Config.Depth.Files, Config.Depth.LastFrame))))  { dataFound = false; return false; }

				return true;
			}
		}
		private bool? dataFound = null; // for caching the result of the DataFound code above

		/// <summary>Indicates if the cached video frames are found in the expected place.</summary>
		public bool CacheFound
		{
			get
			{
				// look up cached value to avoid unnecessary disk accesses
				if (cacheFound != null)
					return (bool)cacheFound;

				cacheFound = true;

				// check if the first and last files of the colour and depth videos exist
				if (!File.Exists(Path.Combine(Config.OutputDir, string.Format(@"processed\video-{0:D4}.raw", Config.Video.FirstFrame)))) { cacheFound = false; return false; }
				if (!File.Exists(Path.Combine(Config.OutputDir, string.Format(@"processed\video-{0:D4}.raw", Config.Video.LastFrame))))  { cacheFound = false; return false; }
				if (!File.Exists(Path.Combine(Config.OutputDir, string.Format(@"processed\depth-{0:D4}.raw", Config.Depth.FirstFrame)))) { cacheFound = false; return false; }
				if (!File.Exists(Path.Combine(Config.OutputDir, string.Format(@"processed\depth-{0:D4}.raw", Config.Depth.LastFrame))))  { cacheFound = false; return false; }

				return true;
			}
		}
		private bool? cacheFound = null; // for caching the result of the CacheFound code above

		public string Dataset
		{
			get { return Config.DatasetName; }
		}

		public override string ToString()
		{
			if (DataFound == false)
				return string.Format("{0} (video not found)", Dataset);
			else if (CacheFound)
				return string.Format("{0} ({1} frames, cached)", Dataset, Frames);
			else
				return string.Format("{0} ({1} frames)", Dataset, Frames);
		}
	}
}
