﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: RendererProgram.cs 553 2012-05-06 19:24:27Z cr333 $
*/
using System;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Security.Permissions;
using System.Threading;
using System.Windows.Forms;
using OpenTK;
using RGBZcamera.Library;
using RGBZcamera.Library.AbstractedLines;
using RGBZcamera.Library.Config;
using RGBZcamera.Library.FillIn;
using RGBZcamera.Library.Hatching;
using RGBZcamera.Library.WinnemoellerAbstraction;
using RGBZcamera.Library.XToon;
using VideoLib;
using VideoLib.Filters;
using VideoLib.Frames;
using VideoLib.LibAv;
using VideoLib.OpenGL;
using VideoLib.OpenGL.OpticalFlow;
using VideoLib.Scheduling;
using VideoLib.Sinks;
using VideoLib.Sources;
using VideoLib.Utils;
using OpenGL = OpenTK.Graphics.OpenGL;
using VGL = VideoLib.OpenGL;

namespace RGBZcamera.Renderer
{
	static class RendererProgram
	{
		public static Topology topology;
		public static Scheduler scheduler;

		static public void Start(RendererForm form, RendererSettings settings)
		{
			topology = Topology.MainInstance;
			scheduler = new ModularScheduler(topology, new OpenGLWorkQueue(), new StatusWorkQueue(), new SharedWorkQueue(4)) { Profiling = true };

//---- load dataset information from config file --------------------------------------------------

			var enableFillIn = true;
			var enableFilter = true;

			// validate settings
			if (settings.ToCache || settings.SaveRenderings) settings.Loop = false;

			var config = settings.Config;

			OutputPin colourTex = null, pointsTex = null, biFiVideoTex = null;
			OpticalFlowFilter flow = null;

			if (settings.ToCache || !settings.FromCache)
			{

//---- get input videos ---------------------------------------------------------------------------

				//// live streaming setup
				//var videoSource = new FlyCaptureSource() { Width = 1024, CaptureFormat = Formats.Yuv422, Framerate = Framerates._7_5, Strobe = true };
				//var sr4000 = new SwissRanger4000Source() { AutoExposure = false, IntegrationTime = 3, Coordinates = false, ConfidenceMap = false, Triggered = true };
				//var distanceSource = sr4000["depth"];

				// config.ini-based
				var distanceSource = new RawSource(config.Depth.Files) { FPS = 15, FirstNumber = config.Depth.FirstFrame, LastNumber = config.Depth.LastFrame, Loop = settings.Loop };

				Node videoSource;
				if (config.Video.Files.ToLower().EndsWith(".raw"))
					videoSource = new RawSource(          config.Video.Files) { FPS = 15, FirstNumber = config.Video.FirstFrame, LastNumber = config.Video.LastFrame, Loop = settings.Loop };
				else
					videoSource = new ImageSequenceSource(config.Video.Files) { FPS = 15, FirstNumber = config.Video.FirstFrame, LastNumber = config.Video.LastFrame, Loop = settings.Loop };

				//// single frame of hand for fill-in figures
				//var videoSource    = new RawSource(config.Video.Files)  { FPS = 5, FirstNumber = 663, LastNumber = 663, Loop = settings.Loop };
				//var distanceSource = new RawSource(config.Depth.Files) { FPS = 3, FirstNumber = 653, LastNumber = 653, Loop = settings.Loop };

				//// single frame of tableA for fill-in comparison
				//var videoSource    = new RawSource(config.Video.Files)  { FPS = 5, FirstNumber = 406, LastNumber = 406, Loop = settings.Loop };
				//var distanceSource = new RawSource(config.Depth.Files) { FPS = 5, FirstNumber = 408, LastNumber = 408, Loop = settings.Loop };

				//// save to disk
				//topology.Connect(videoSource, new ImageSequenceSink(Path.Combine(config.OutputDir, @"input\video-{0:D4}.png")));
				////topology.Connect(distanceSource, new ConvertFilter(PixelFormat.Grey8), new ImageSequenceSink(Path.Combine(config.OutputDir, @"0-input\distance-{0:D4}.png"))); // raw input distance map (scaled to fit 5m into 0..1)
				//var inputPoints = topology.Connect(distanceSource, new ConvertFilter(PixelFormat.Float), new DistanceImageToPointCloud { Intrinsics = config.SR4000.IntrParams });
				//topology.Connect(inputPoints, new TextureInput(), new PointCloudToDistanceMap(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"0-input\distance-{0:D4}.png")));
				//topology.Connect(inputPoints, new OffExportSink(Path.Combine(config.OutputDir, @"0-input\mesh-{0:D4}.off")));

//---- undistort videos and pre-process distance map ----------------------------------------------

				// undistort video
				var videoTex = topology.Connect(videoSource,
					new TextureInput(),
					new UndistortFilter() { Parameters = config.Video.IntrParams });
				var video = topology.Connect(videoTex, new TextureOutput(PixelFormat.BGR32));

				// undistort distance
				float depthScale = 1.0f;
				if (config.ReadSetting("General", "SensorType", SensorType.FleaSR) == (int)SensorType.Kinect) depthScale = 13.1072f; // specific scale factor for Kinect
				var distanceTex = topology.Connect(distanceSource,
					new TextureInput() { InternalFormat = OpenGL::PixelInternalFormat.R32f },
					new LinearTransform() { Factor = new Vector4(depthScale, depthScale, depthScale, depthScale) },
					new MedianFillIn(), new MedianFillIn(), new MedianFillIn(),
					new PixelSearchReplace(),
					new UndistortFilter() { Parameters = config.Depth.IntrParams });
				var distance = topology.Connect(distanceTex, new TextureOutput(PixelFormat.Float));

				//// save to disk
				//topology.Connect(video, new ImageSequenceSink(Path.Combine(config.OutputDir, @"undistorted\video-{0:D4}.png")));
				////topology.Connect(distance, new ConvertFilter(PixelFormat.Grey8), new ImageSequenceSink(Path.Combine(config.OutputDir, @"1-undistorted\distance-{0:D4}.png"))); // raw input depth map undistorted (scaled to fit 5m into 0..1)

//---- reproject distance map ---------------------------------------------------------------------

				// reproject distance map
				var points = new DistanceImageToPointCloud() { Intrinsics = config.Depth.IntrParams, };
				config.ReadSettings(points, "Reprojection", "MapType");
				//topology.Connect(points, new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA }, new ImageSequenceSink(Path.Combine(config.OutputDir, @"1-undistorted\distance-{0:D4}.png")));
				//topology.Connect(points, new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA }, new OffExportSink(Path.Combine(config.OutputDir, @"1-undistorted\mesh-{0:D4}.off")));

				var reproj = new ReprojectionNode()
				{
					SourceExtrinsics = config.Depth.ExtrParams,
					SourceIntrinsics = config.Depth.IntrParams,
					TargetExtrinsics = config.Video.ExtrParams,
					TargetIntrinsics = config.Video.IntrParams,
				};
				config.ReadSettings(reproj, "Reprojection", "CropNear", "CropFar", "CleanupMode", "NormalThreshold", "SobelThreshold");

				topology.Connect(videoTex,            reproj["colourIn"]);
				topology.Connect(distanceTex, points, reproj["pointsIn"]);

				colourTex = reproj.OutputPins["colourOut"];
				pointsTex = reproj.OutputPins["pointsOut"];

				//// save to disk
				//// pass 1: just reprojected/aligned, no geometry invalidation ('aligned' mesh)
				//reproj.CleanupMode = DepthCleanupMode.Disabled; // no geometry invalidation
				//topology.Connect(colourTex, new TextureOutput(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"2-aligned\video-{0:D4}.png")));
				//topology.Connect(pointsTex, new PointCloudToDistanceMap(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"2-aligned\distance-{0:D4}.png"))); // tableA: 0.5z-0.5
				//topology.Connect(pointsTex, new TextureOutput(PixelFormat.FloatRGBA), new OffExportSink(Path.Combine(config.OutputDir, @"2-aligned\mesh-{0:D4}.off")));
				// pass 2: with geometry invalidation
				//topology.Connect(pointsTex, new PointCloudToDistanceMap(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"3-invalidated-distance\distance-{0:D4}.png")));
				//topology.Connect(pointsTex, new TextureOutput(PixelFormat.FloatRGBA), new OffExportSink(Path.Combine(config.OutputDir, @"3-invalidated-mesh\mesh-{0:D4}.off")));

//---- inpainting of invalid areas ----------------------------------------------------------------

				// bilaterally filter the input video
				biFiVideoTex = topology.Connect(colourTex, new BilateralFilter() { SpatialSigma = 3, RangeSigma = 0.1f }).OutputPins["out"];

				//// OpenGL rendering passes are executed asynchronously by the GPU and OpenGL nodes hence can
				//// return before their last rendering pass has completed. This causes the reported runtime of
				//// a node to be lower than it actually is, as part of it is 'moved' to the next node in the
				//// pipeline, which will wait for the texture rendered by the previous node to become available.
				//// For measuring runtimes more accurately, we hence insert texture-pass-through nodes before
				//// and after the multi-resolution geometry fill-in. The second, longer-executing node's average
				//// runtime will then be added to the multi-res runtime.
				//pointsTex = topology.Connect(pointsTex, new PassThroughFilter()).OutputPins["out"];

				var fillin = new MultiResolutionFillIn() { Enabled = enableFillIn };
				config.ReadSettings(fillin, "FillIn", "Factor", "Levels", "RangeSigma", "SpatialSigma");
				topology.Connect(biFiVideoTex, fillin["colour"]);
				topology.Connect(pointsTex,    fillin["points"]);
				pointsTex = fillin.OutputPins["out"];

				//// for timing purposes (see above for explanation)
				//pointsTex = topology.Connect(pointsTex, new PassThroughFilter()).OutputPins["out"];

				//// settings for multi-res fill-in figure in paper (use hand frame 663)
				//reproj.CleanupMode = DepthCleanupMode.PixelNormalThresholding;
				//reproj.NormalThreshold = 0.6f;
				//fillin.Factor = 3;
				//fillin.Levels = 3;
				//fillin.RangeSigma = 0.02f;
				//fillin.SpatialSigma = 3;

				//// save to disk
				//topology.Connect(pointsTex, new PointCloudToDistanceMap(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"4-filled-in\distance-{0:D4}.png")));
				//topology.Connect(pointsTex, new LinearTransform() { Factor = new Vector4(1, 1, 1, 0.5f), Offset = new Vector4(0, 0, 0, -0.5f) }, new PointCloudToDistanceMap(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"4-filled-in\distance-{0:D4}.png")));
				//topology.Connect(pointsTex, new TextureOutput(PixelFormat.FloatRGBA), new OffExportSink(Path.Combine(config.OutputDir, @"4-filled-in\mesh-{0:D4}.off")));

//---- spatio-temporal filtering ------------------------------------------------------------------

				// optical flow between video frames
				flow = new OpticalFlowFilter() { CoarseLevel = 6, Alpha = 8, Iterations = 40 };
				config.ReadSettings(flow, "Flow", "Iterations", "JacobiIterations", "FineLevel", "CoarseLevel", "LevelOffset", "Alpha", "Beta");
				var lastColourTex = topology.Connect(colourTex, new DelayFilter(1) { DuplicateFirst = true });
				topology.Connect(lastColourTex, flow["in0"]);
				topology.Connect(colourTex,     flow["in1"]);

				//// visualise flow
				//var flowVis = new FlowVisualiser() { Width = 1024, Height = 768, GridSize = 8, ArrowScale = 1, ArrowColor = System.Drawing.Color.Brown };
				//topology.Connect(videoTex, flowVis["image"]);
				//topology.Connect(flow,     flowVis["flow"]);
				//var vis = topology.Connect(flowVis, new TextureOutput());
				//topology.ConnectFan(vis, new VideoLib.OpenGL.VideoDisplay(form.videoControl6, "Optical flow arrows"));
				////topology.Connect(vis, new ImageSequenceSink(Path.Combine(config.OutputDir, @"flow-vis\{0:D4}.png")));

				//// simple flow-based exponential averaging
				//var averageTex = new FlowAverager();
				//topology.Connect(pointsTex, new TextureInput() { InternalFormat = OpenGL::PixelInternalFormat.Rgba32f }, averageTex);
				//topology.Connect(flow, averageTex["flow"]);
				//var average = topology.Connect(averageTex, new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA });
				//pointsTex = average.OutputPins["out"];

				// spatio-temporal bilateral filter
				var spatiotemporal = new SpatioTemporalFilter() { Enabled = enableFilter };
				config.ReadSettings(spatiotemporal, "STF", "SpatialSigma", "ColourSigma", "DistanceSigma", "FalloffWeight", "FlowSigma");
				topology.Connect(biFiVideoTex, spatiotemporal["currentVideo"]);
				topology.Connect(biFiVideoTex, new DelayFilter(1) { DuplicateFirst = true }, spatiotemporal["previousVideo"]);
				topology.Connect(flow, spatiotemporal["flow"]);
				topology.Connect(pointsTex, spatiotemporal["distance"]);
				pointsTex = spatiotemporal.OutputPins["out"];

				//// save to disk ('filtered' mesh)
				//topology.Connect(pointsTex, new ConvertFilter(PixelFormat.Byte) { Offset = 0.5f }, new LibAvConvert(PixelFormat.BGR32), new ImageSequenceSink(Path.Combine(config.OutputDir, @"filtered\{0:D4}.png")));
				//topology.Connect(pointsTex, new PointCloudToDistanceMap(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"5-filtered\distance-{0:D4}.png")));
				//topology.Connect(pointsTex, new PointCloudToDistanceMap() { Format = PixelFormat.Float }, new RawSink(Path.Combine(config.OutputDir, @"5-filtered\distance-{0:D4}.raw")));
				//topology.Connect(pointsTex, new TextureOutput(PixelFormat.FloatRGBA), new OffExportSink(Path.Combine(config.OutputDir, @"off\mesh-{0:D4}.off"))); // occasionally causes colour image to be black which breaks the filter and produces poor results

				//// save to disk ('spatial' mesh)
				//spatiotemporal.SpatialSigma  = 8;
				//spatiotemporal.FalloffWeight = 0;
				//spatiotemporal.FlowSigma     = 10000;
				//topology.Connect(pointsTex, new PointCloudToDistanceMap() { Format = PixelFormat.Float }, new RawSink(Path.Combine(config.OutputDir, @"9-spatial\distance-{0:D4}.raw")));
				//topology.Connect(pointsTex, new TextureOutput(PixelFormat.FloatRGBA), new OffExportSink(Path.Combine(config.OutputDir, @"off\mesh-{0:D4}.off"))); // occasionally causes colour image to be black which breaks the filter and produces poor results


//---- display intermediate results ---------------------------------------------------------------

				if (settings.DisplayDebugOutputs)
				{
					var sync = new SyncFilter(5);
					topology.Connect(sync["out0"], new VGL.VideoDisplay(form.videoControl1, "Video input"));
					topology.Connect(sync["out1"], new VGL.VideoDisplay(form.videoControl2, "Distance input"));
					topology.Connect(sync["out2"], new VGL.VideoDisplay(form.videoControl3, "Cropped video"));
					topology.Connect(sync["out3"], new VGL.VideoDisplay(form.videoControl4, "Reprojected distance"));
					topology.Connect(sync["out4"], new VGL.VideoDisplay(form.videoControl5, "Filtered distance"));

					// inputs
					topology.Connect(videoSource, sync["in0"]);
					topology.Connect(distance, sync["in1"]);

					// reprojected video & distance map
					topology.Connect(reproj["colourOut"], new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA }, sync["in2"]);
					topology.Connect(reproj["pointsOut"], new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA }, sync["in3"]);

					//---- show filtered depth ----
					topology.Connect(pointsTex, new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA }, sync["in4"]);
				}
				else
				{
					// need to read out final result, for reliable run time measurements
					topology.Connect(pointsTex, new TextureOutput());
				}

//==== save/load result to/from disk ==============================================================

				// save processed data to disk
				if (settings.ToCache)
				{
					topology.Connect(colourTex, new TextureOutput() { TargetFormat = PixelFormat.BGR24     }, new RawSink(Path.Combine(config.OutputDir, @"processed\video-{0:D4}.raw"))); // TODO: test
					topology.Connect(pointsTex, new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA }, new RawSink(Path.Combine(config.OutputDir, @"processed\depth-{0:D4}.raw")));
				}
			}
			else // load processed data from disk
			{
				var colour = new RawSource(Path.Combine(config.OutputDir, @"processed\video-{0:D4}.raw")) { FPS = 15, FirstNumber = config.Video.FirstFrame,  LastNumber = config.Video.LastFrame,  Loop = settings.Loop }.OutputPins["out"];
				var points = new RawSource(Path.Combine(config.OutputDir, @"processed\depth-{0:D4}.raw")) { FPS = 15, FirstNumber = config.Depth.FirstFrame, LastNumber = config.Depth.LastFrame, Loop = settings.Loop };

				colourTex = topology.Connect(colour, new TextureInput() { InternalFormat = OpenGL::PixelInternalFormat.Rgba8,   Wrap = OpenGL::TextureWrapMode.ClampToEdge }).OutputPins["out"];
				pointsTex = topology.Connect(points, new TextureInput() { InternalFormat = OpenGL::PixelInternalFormat.Rgba32f, Wrap = OpenGL::TextureWrapMode.ClampToEdge }).OutputPins["out"];

				//// save to disk
				//topology.Connect(pointsTex, new PointCloudToDistanceMap(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"5-filtered\distance-{0:D4}.png")));
				//topology.Connect(pointsTex, new TextureOutput(PixelFormat.FloatRGBA), new OffExportSink(Path.Combine(config.OutputDir, @"5-filtered\mesh-{0:D4}.off")));
			}

//==== Eendering frontend =========================================================================

			SwitchFilter select = null;

			// case split to pass on the optical flow if it is computed
			if (settings.FromCache) select = new SwitchFilter(9, "colour", "points");
			else
			{
				select = new SwitchFilter(9, "colour", "flow", "points");
				topology.Connect(flow, select["flowIn"]);
			}
			topology.Connect(colourTex, select["colourIn"]);
			topology.Connect(pointsTex, select["pointsIn"]);
			select.SelectedOutput = 2;
			//select.MultipleOutputs = true; select.SelectedOutput = 126; // all results but mesh
			form.StyleSelector = select;

			var display = new VGL.VideoDisplay(form.videoControl6);

			// save input video
			//if (settings.SaveRenderings) topology.Connect(colourTex, new TextureOutput(), new ImageSequenceSink(Path.Combine(config.ResultsDir, @"input\video-{0:D4}.png")));
			//if (settings.SaveRenderings) topology.Connect(biFiVideoTex, new TextureOutput(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"input\bilateral-{0:D4}.png")));

			// 0. Render filtered depth as mesh
			{
				var renderedMeshTex = new MeshRenderer() { Size = new System.Drawing.Size(640, 480), Eye = new OpenTK.Vector3(-0.2f, 0.2f, 0), LookAt = new OpenTK.Vector3(0.1f, 0.0f, 1) }; // looking from top-right
				//var renderer = new MeshRenderer() { Size = new System.Drawing.Size(640, 480), Eye = new OpenTK.Vector3(0.4f, 0.4f, 0), LookAt = new OpenTK.Vector3(-0.2f, 0.0f, 1) }; // looking from top-left
				var renderedMesh = topology.Connect(select["pointsOut0"], renderedMeshTex, new TextureOutput());

				if (settings.DisplayRenderings) topology.Connect(renderedMesh, display);
				if (settings.SaveRenderings)    topology.Connect(renderedMesh, new ImageSequenceSink(Path.Combine(config.OutputDir, @"mesh\{0:D4}-mesh.png")));
			}

			// 1. Debug view, e.g. alignment of video and depth data, or diffuse rendering
			{
				var debug = new ToneRendererNode() { ToneOutput = ToneOutput.Diffuse };
				topology.Connect(select["colourOut1"], debug["colour"]);
				topology.Connect(select["pointsOut1"], debug["depth"]);

				if (settings.DisplayRenderings) topology.Connect(debug, new TextureOutput(PixelFormat.FloatRGBA), display);
				if (settings.SaveRenderings)    topology.Connect(debug, new TextureOutput(), new ImageSequenceSink(Path.Combine(config.ResultsDir, @"diffuse\diffuse-{0:D4}.png")));
			}

			// 2. Line drawings via abstracted shading (Lee et al. 2007)
			{
				var abstractedTex = new AbstractedLinesFilter();
				config.ReadSettings(abstractedTex, "Abstraction", "LightDirection", "IsPosition", "Ambient", "Diffuse", "Specular", "Shininess", "ToonStep", "Sigma", "LineWidth", "NumSteps", "Beta", "RidgesD", "RidgesS", "Valleys", "BlendMode", "BlendFactor", "Saturation");
				topology.Connect(select["colourOut2"], new BilateralFilter() { SpatialSigma = 3, RangeSigma = 0.1f }, abstractedTex["colour"]);
				topology.Connect(select["pointsOut2"],                                                                abstractedTex["points"]);
				var abstracted = topology.Connect(abstractedTex, new TextureOutput());

				if (settings.DisplayRenderings) topology.Connect(abstracted, display);
				if (settings.SaveRenderings)    topology.Connect(abstracted, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"abstraction\abstracted-{0:D4}.png")));
			}

			// 3. Relighting
			{
				var relitTex = new Relighter();
				config.ReadSettings(relitTex, "General", "BackgroundPlane", "PlaneThreshold");
				config.ReadSettings(relitTex, "Relighting", "LightingCoefficients", "EnvironmentMap", "ScaleAlbedo", "Rotation", "OutputMatte");
				//relitTex.EnvironmentMap = EnvironmentMap.Beach;
				topology.Connect(select["colourOut3"], relitTex["colour"]);
				topology.Connect(select["pointsOut3"], relitTex["points"]);
				var relit = topology.Connect(relitTex, new TextureOutput());

				if (settings.DisplayRenderings) topology.Connect(relit, display);
				if (settings.SaveRenderings)    topology.Connect(relit, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"relighting\" + relitTex.EnvironmentMap.ToString() + "-{0:D4}.png")));
				//if (settings.SaveRenderings)    topology.Connect(relit, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"relighting\matte-{0:D4}.png")));
			}

			// 4. X-Toon
			{
				var xtoonTex = new XToonShader() { AttributeMap = AttributeMapType.Highlights, Look = Look.SilhouetteSmoothing, SequenceNumberDelta = config.Video.FirstFrame - config.Depth.FirstFrame };
				config.ReadSettings(xtoonTex, "XToon", "AttributeMap", "Look", "DepthRange", "LightDirection", "IsPosition", "Diffuse", "Specular", "Shininess");
				var xtoon = topology.Connect(select["pointsOut4"], xtoonTex, new TextureOutput());

				if (settings.DisplayRenderings) topology.Connect(xtoonTex, display);
				if (settings.SaveRenderings)    topology.Connect(xtoon, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"xtoon\Highlights-SilhouetteSmoothing-{0:D4}.png")));
			}

			// 5. Hatching
			{
				var strokesTex = new HatchingFilter();
				config.ReadSettings(strokesTex, "General", "BackgroundPlane", "PlaneThreshold");
				config.ReadSettings(strokesTex, "Strokes", "AddThreshold", "DeleteThreshold", "CoverageSize", "SpriteSize", "TargetCoverage", "BlackOrWhite", "Attenuation", "LightDirection", "JitterPosition", "JitterRotation", "OrientByColour", "SmoothOrientation");
				topology.Connect(select["colourOut5"], strokesTex["colour"]);
				topology.Connect(select["pointsOut5"], strokesTex["points"]);

				if (settings.FromCache && !settings.ToCache)
				{
					// need to compute flow if loading from disk
					flow = new OpticalFlowFilter() { CoarseLevel = 6, Alpha = 8, Iterations = 40 };
					config.ReadSettings(flow, "Flow", "Iterations", "JacobiIterations", "FineLevel", "CoarseLevel", "LevelOffset", "Alpha", "Beta");
					var lastColourTex = topology.Connect(select["colourOut5"], new DelayFilter(1) { DuplicateFirst = true });
					topology.Connect(lastColourTex,        flow["in0"]);
					topology.Connect(select["colourOut5"], flow["in1"]);
					topology.Connect(flow,                 strokesTex["flow"]);
				}
				else
				{
					// reuse existing flow
					topology.Connect(select["flowOut5"], strokesTex["flow"]);
				}

				var strokes = topology.Connect(strokesTex, new TextureOutput());
				if (settings.DisplayRenderings) topology.Connect(strokes, display);
				if (settings.SaveRenderings)    topology.Connect(strokes, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"strokes\strokes-{0:D4}.png")));
			}

			// 6. Real-time video abstraction (Winnemöller et al. 2006)
			{
				var winnemoellerTex = new WinnemoellerAbstraction();
				config.ReadSettings(winnemoellerTex, "Winnemoeller", "SigmaR", "SigmaD", "N_E", "N_B", "Saturation", "SigmaE", "Q", "LambdaDelta", "OmegaDelta", "LambdaPhi", "OmegaPhi", "PhiE", "Tau");
				var winnemoeller = topology.Connect(select["colourOut6"], winnemoellerTex, new TextureOutput());

				if (settings.DisplayRenderings) topology.Connect(winnemoeller, display);
				if (settings.SaveRenderings)    topology.Connect(winnemoeller, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"winnemoeller\winnemoeller-{0:D4}.png")));
			}

			// 7. Background subtraction
			{
				var backSubTex = new BackgroundSubtraction();
				config.ReadSettings(backSubTex, "General", "BackgroundPlane", "PlaneThreshold");
				topology.Connect(select["colourOut7"], backSubTex["colour"]);
				topology.Connect(select["pointsOut7"], backSubTex["points"]);
				var backSub = topology.Connect(backSubTex, new TextureOutput());

				if (settings.DisplayRenderings) topology.Connect(backSub, display);
				if (settings.SaveRenderings && !backSubTex.ShowMatte) topology.Connect(backSub, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"background-segmentation\foreground-{0:D4}.png")));
				if (settings.SaveRenderings &&  backSubTex.ShowMatte) topology.Connect(backSub, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"background-segmentation\matte-{0:D4}.png")));
			}

			// 8. Stereo rendering
			{
				var stereoWarp = new StereoWarper()
				{
					// same settings as ReprojectionNode
					SourceExtrinsics = config.Depth.ExtrParams,
					SourceIntrinsics = config.Depth.IntrParams,
					TargetExtrinsics = config.Video.ExtrParams,
					TargetIntrinsics = config.Video.IntrParams,
				};

				// read settings and connect up stereo warper
				config.ReadSettings(stereoWarp, "General", "SensorType");
				config.ReadSettings(stereoWarp, "Stereo", "InterOcularDistance", "ScreenDistance");
				topology.Connect(select["colourOut8"], stereoWarp["colourIn"]);
				topology.Connect(select["pointsOut8"], stereoWarp["pointsIn"]);
				var colourLTex = stereoWarp["colourOutL"];
				var colourRTex = stereoWarp["colourOutR"];
				//stereoWarp.InterOcularDistance = 0.1f; // 'wide' baseline
				//stereoWarp.Shift = 2; // leszek sequences

				//// Geometry abstraction in stereo 3D
				//var abstractedLTex = new AbstractedLinesFilter();
				//var abstractedRTex = new AbstractedLinesFilter();
				//config.ReadSettings(abstractedLTex, "Abstraction", "LightDirection", "IsPosition", "Ambient", "Diffuse", "Specular", "Shininess", "ToonStep", "Sigma", "LineWidth", "NumSteps", "Beta", "RidgesD", "RidgesS", "Valleys", "BlendMode", "BlendFactor", "Saturation");
				//config.ReadSettings(abstractedRTex, "Abstraction", "LightDirection", "IsPosition", "Ambient", "Diffuse", "Specular", "Shininess", "ToonStep", "Sigma", "LineWidth", "NumSteps", "Beta", "RidgesD", "RidgesS", "Valleys", "BlendMode", "BlendFactor", "Saturation");
				//topology.Connect(stereoWarp["colourOutL"], new BilateralFilter() { SpatialSigma = 3, RangeSigma = 0.1f }, abstractedLTex["colour"]);
				//topology.Connect(stereoWarp["pointsOutL"], abstractedLTex["points"]);
				//topology.Connect(stereoWarp["colourOutR"], new BilateralFilter() { SpatialSigma = 3, RangeSigma = 0.1f }, abstractedRTex["colour"]);
				//topology.Connect(stereoWarp["pointsOutR"], abstractedRTex["points"]);
				//colourLTex = abstractedLTex["out"];
				//colourRTex = abstractedRTex["out"];

				//// Video relighting in stereo 3D
				//var relitLTex = new Relighter() { EnvironmentMap = EnvironmentMap.Beach };
				//var relitRTex = new Relighter() { EnvironmentMap = EnvironmentMap.Beach };
				//config.ReadSettings(relitLTex, "General", "BackgroundPlane", "PlaneThreshold");
				//config.ReadSettings(relitRTex, "General", "BackgroundPlane", "PlaneThreshold");
				//config.ReadSettings(relitLTex, "Relighting", "LightingCoefficients", "EnvironmentMap", "ScaleAlbedo", "Rotation", "OutputMatte");
				//config.ReadSettings(relitRTex, "Relighting", "LightingCoefficients", "EnvironmentMap", "ScaleAlbedo", "Rotation", "OutputMatte");
				//topology.Connect(stereoWarp["colourOutL"], relitLTex["colour"]);
				//topology.Connect(stereoWarp["pointsOutL"], relitLTex["points"]);
				//topology.Connect(stereoWarp["colourOutR"], relitRTex["colour"]);
				//topology.Connect(stereoWarp["pointsOutR"], relitRTex["points"]);
				//colourLTex = relitLTex["out"];
				//colourRTex = relitRTex["out"];

				//// Winnemöller in stereo 3D
				//var winnemoellerLTex = new WinnemoellerAbstraction();
				//var winnemoellerRTex = new WinnemoellerAbstraction();
				//config.ReadSettings(winnemoellerLTex, "Winnemoeller", "SigmaR", "SigmaD", "N_E", "N_B", "Saturation", "SigmaE", "Q", "LambdaDelta", "OmegaDelta", "LambdaPhi", "OmegaPhi", "PhiE", "Tau");
				//config.ReadSettings(winnemoellerRTex, "Winnemoeller", "SigmaR", "SigmaD", "N_E", "N_B", "Saturation", "SigmaE", "Q", "LambdaDelta", "OmegaDelta", "LambdaPhi", "OmegaPhi", "PhiE", "Tau");
				//topology.Connect(stereoWarp["colourOutL"], winnemoellerLTex);
				//topology.Connect(stereoWarp["colourOutR"], winnemoellerRTex);
				//colourLTex = winnemoellerLTex["out"];
				//colourRTex = winnemoellerRTex["out"];

				// render the stereo image as an anaglyph image
				var anaglyph = new AnaglyphFilter(AnaglyphMethod.HalfColour);
				var colourL = topology.Connect(colourLTex, new TextureOutput());
				var colourR = topology.Connect(colourRTex, new TextureOutput());
				topology.Connect(colourL, anaglyph["inL"]);
				topology.Connect(colourR, anaglyph["inR"]);

				if (settings.DisplayRenderings) topology.Connect(anaglyph, display);
				//if (settings.DisplayRenderings) topology.Connect(colourR, display);
				if (settings.SaveRenderings)
				{
					topology.Connect(colourL,  new ImageSequenceSink(Path.Combine(config.ResultsDir, @"stereo\stereo-L-{0:D4}.png")));
					topology.Connect(colourR,  new ImageSequenceSink(Path.Combine(config.ResultsDir, @"stereo\stereo-R-{0:D4}.png")));
					//topology.Connect(anaglyph, new ImageSequenceSink(Path.Combine(config.ResultsDir, @"stereo\stereo-A-{0:D4}.png")));
				}
			}

			////// save to disk
			////topology.Connect(pointsTex, renderer, new ConvertFilter(PixelFormat.BGR24), new ImageSequenceSink(Path.Combine(config.OutputDir, @"2010-11-02\spatiotemporal\{0:D4}-mesh-w=0.0-s6.png")));
			//var diffuse = new ToneRendererNode { ToneOutput = ToneOutput.Diffuse };
			//topology.Connect(biFiVideoTex, diffuse["colour"]);
			//topology.Connect(filteredPointsTex, diffuse["depth"]);
			//topology.Connect(diffuse, new TextureOutput(), new ConvertFilter(PixelFormat.BGR24), new ImageSequenceSink(Path.Combine(config.OutputDir, @"new-sequences\{0:D4}-" + dataset + "-diffuse.png")));
			//topology.Connect(ridge, new TextureOutput(), new ConvertFilter(PixelFormat.BGR24), new ImageSequenceSink(Path.Combine(config.OutputDir, @"new-sequences\{0:D4}-" + dataset + "-toon.png")));
			//topology.Connect(alignment, new TextureOutput(), new ConvertFilter(PixelFormat.BGR24), new ImageSequenceSink(Path.Combine(config.OutputDir, @"flow-filter\{0:D4}-07-Gaussian-motion-flow-sigma.png")));

//---- property editor logic ----------------------------------------------------------------------

			// Pass list of processing nodes to the form to be able to change settings.
			form.BeginInvoke(new Action(delegate()
			{
				form.nodeComboBox.Items.Clear();
				form.nodeComboBox.Items.AddRange(topology.Nodes.OrderBy(x => x.Name).ToArray());
				form.SelectStyleNode();
			}));

			// HACK: force garbage collection
			var gcEnforcer = new Thread(ForceGC);
			gcEnforcer.IsBackground = true;
			gcEnforcer.Start();

//---- start the scheduler ------------------------------------------------------------------------

			//// export topology as DOT graph
			//using (var dotFile = new StreamWriter("topology.dot"))
			//    dotFile.Write(topology.ExportToDotGraph());
			//return;

			scheduler.Setup();
			scheduler.StartWorkers(4);
			scheduler.Join();
		}

		static RendererForm form;

		/// <summary>The main entry point for the application.</summary>
		[STAThread]
		static void Main()
		{
			Application.EnableVisualStyles();
			Application.SetCompatibleTextRenderingDefault(false);
#if !DEBUG
			RegisterUnhandledExceptionEventHandler();

			// force all Windows Forms errors to go through our handler
			Application.SetUnhandledExceptionMode(UnhandledExceptionMode.CatchException);
			Application.ThreadException += new ThreadExceptionEventHandler(Application_ThreadException);
#endif
			Application.Run(form = new RendererForm());
		}

		public static void ForceGC()
		{
			while (true)
			{
				GC.Collect();
				Thread.Sleep(2000);
			}
		}
		
		#region Exception handlers
//=================================================================================================

		[SecurityPermission(SecurityAction.Demand, Flags = SecurityPermissionFlag.ControlAppDomain)]
		public static void RegisterUnhandledExceptionEventHandler()
		{
			AppDomain currentDomain = AppDomain.CurrentDomain;
			currentDomain.UnhandledException += new UnhandledExceptionEventHandler(AppDomain_UnhandledException);
		}

		static void Application_ThreadException(object sender, ThreadExceptionEventArgs args)
		{
			Exception e = args.Exception;
			MessageBox.Show(
				String.Format("{0}: {1}\n\n{2}\n\n(Application.ThreadException)", e.GetType().Name, e.Message, e),
				String.Format("Error: {0}", e.GetType()), MessageBoxButtons.OK, MessageBoxIcon.Error);
			form.BeginInvoke(new Action(delegate() { form.Close(); }));
		}

		static void AppDomain_UnhandledException(object sender, UnhandledExceptionEventArgs args)
		{
			Exception e = (Exception)args.ExceptionObject;
			MessageBox.Show(
				String.Format("{0}: {1}\n\n{2}\n\n(AppDomain.UnhandledException)", e.GetType().Name, e.Message, e),
				String.Format("Error: {0}", e.GetType()), MessageBoxButtons.OK, MessageBoxIcon.Error);
			form.BeginInvoke(new Action(delegate() { form.Close(); }));
		}

//=================================================================================================
		#endregion // Exception handlers
	}
}
