/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

This code is based on original code by Chenglei Wu, MPI Informatik (Nov 2010).

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: LightingEstimationStep2.cpp 539 2012-04-04 15:42:03Z cr333 $
*/
#pragma comment (lib, "highgui210.lib")
#pragma comment (lib, "cv210.lib")
#pragma comment (lib, "cxcore210.lib")

#include <opencv/cv.h>
#include <opencv/cxcore.h>
#include <opencv/highgui.h>
#include <vector>
#include <stdio.h>
#include <stdlib.h>
#include <sys/timeb.h>
#include <time.h>

#define PI 3.1415926

#define NUMORD 3

using namespace std;

class CVec3f
{
public:
	float v[3];

	float& operator[] (int i) { return v[i]; }
};

static float g_scalefactor[18];
static float g_tmpkn[9] = { 0.8862f, 1.0233f, 0.4954f, 0.0f , -0.1108f, 0.0f, 0.0499f, 0.0f, -0.0285f };

template <typename T> T clamp(T x, T minVal, T maxVal)
{
	if(x < minVal) return minVal;
	else if(x > maxVal) return maxVal;
	else return x;
}

/// UNIX timestamp: seconds from 1970-01-01 00:00:00 (UTC)
double mtime(void)
{
	_timeb ts;

#ifdef _MSC_VER
	_ftime_s(&ts); // special code to please MS VC++
#else // _MSC_VER
	_ftime(&ts);
#endif // _MSC_VER

	return ts.time + ts.millitm / 1000.0;
}

int GetSHval(CVec3f m_normal, vector<float>& vec_shval)
{
	if(vec_shval.size() != NUMORD*NUMORD)
		return -1;

	float px, py, pz;
	px = m_normal[0];
	py = m_normal[1];
	pz = m_normal[2];

	float lenth = (float)sqrt(px * px + py * py + pz * pz);
	float lenth2 = lenth * lenth;
	float lenth4 = lenth2 * lenth2;
	float px2 = px * px;
	float py2 = py * py;
	float pz2 = pz * pz;

	if(!lenth)
		return -1;

	float m_tmpshade = 0.0f;
	vec_shval[0] = 1;
	vec_shval[1] = py / lenth;
	vec_shval[2] = pz / lenth;
	vec_shval[3] = px / lenth;
	vec_shval[4] = px * py / lenth2;
	vec_shval[5] = py * pz / lenth2;
	vec_shval[6] = (-px2 - py2 + 2 * pz2) / lenth2;
	vec_shval[7] = pz * px / lenth2;
	vec_shval[8] = (px2 - py2) / lenth2;

#ifdef HIGH_ORDER
	for(int i=9;i<16;i++)
	{
		vec_shval[i] = 0.0f;
	}

	vec_shval[16] = px*py*(px2-py2)/lenth4;
	vec_shval[17] = (3*px2-py2)*py*pz/lenth4;
	vec_shval[18] = px*py*(7*pz2-lenth2)/lenth4;
	vec_shval[19] = py*pz*(7*pz2-3*lenth2)/lenth4;
	vec_shval[20] = (35*pz2*pz2-30*pz2*lenth2+3*lenth4)/lenth4;
	vec_shval[21] = px*pz*(7*pz2-3*lenth2)/lenth4;
	vec_shval[22] = (px2-py2)*(7*pz2-lenth2)/lenth4;
	vec_shval[23] = (px2-3*py2)*px*pz/lenth4;
	vec_shval[24] = (px2*(px2-3*py2)-py2*(3*px2-py2))/lenth4;
#endif


	for(int n = 0; n < NUMORD; ++n)
	{
		if(n == 3) continue;

		int tmpbias = 0;
		if(n == 4)
			tmpbias = -7;

		int m = 0;
		int m_tmpindtmp = 1 + n * (n + 1) + m - 1;
		vec_shval[m_tmpindtmp] = vec_shval[m_tmpindtmp] * g_tmpkn[n] * g_scalefactor[tmpbias + m_tmpindtmp];

		for(m = 1; m < n + 1; ++m)
		{
			m_tmpindtmp = 1 + n * (n + 1) + m - 1;
			vec_shval[m_tmpindtmp] = vec_shval[m_tmpindtmp] * g_tmpkn[n] * g_scalefactor[tmpbias + m_tmpindtmp];

			m_tmpindtmp = 1 + n * (n + 1) - m - 1;
			vec_shval[m_tmpindtmp] = vec_shval[m_tmpindtmp] * g_tmpkn[n] * g_scalefactor[tmpbias + m_tmpindtmp];
		}
	}

	return 1;
}

// read colours and normals from text file
int ReadFrame(const char* filename, vector<CvPoint>& vec_pos, vector<CVec3f>& vec_normal, vector<CVec3f>& vec_color)
{
	printf("Reading in \"%s\" ... ", filename);
	FILE* pfile = fopen(filename, "r");

	while(!feof(pfile))
	{
		CvPoint pt_tmp;
		CVec3f  nor_tmp;
		CVec3f  col_tmp;
		if(8 != fscanf(pfile, "%d %d %f %f %f %f %f %f\n", &pt_tmp.x, &pt_tmp.y, &nor_tmp[0], &nor_tmp[1], &nor_tmp[2], &col_tmp[0], &col_tmp[1], &col_tmp[2]))
		{
			printf("\rReading in \"%s\" ... error while reading in\n", filename);
			return -1;
		}

		vec_pos.push_back(pt_tmp);
		vec_normal.push_back(nor_tmp);
		vec_color.push_back(col_tmp);

		if(vec_pos.size() % 1000 == 0) printf("\rReading in \"%s\" ... %i pixels read", filename, vec_pos.size());
	}

	fclose(pfile);
	printf("\rReading in \"%s\" ... OK.                \n", filename);
	return 1;
}

int AlbedoEstimation(const char* maskimgfile, const vector<CvPoint>& vec_pos, const vector<CVec3f>& vec_normal, vector<CVec3f>& vec_color)
{
	printf("Albedo Estimation ... ");

	vector<CVec3f> vec_lit(NUMORD * NUMORD);
	FILE *plitfile = fopen("litcoef.txt", "r");
	for (int nchal = 0; nchal < 3; nchal++)
		for (int i = 0; i < NUMORD * NUMORD; i++)
			fscanf(plitfile, "%f", &vec_lit[i][nchal]);
	fclose(plitfile);

	int m_flag = 0;
	float m_pream = vec_lit[0][0];
	for(int i = 1; i < 3; i++)
		if(vec_lit[0][i] > m_pream)
			m_flag = i;

	//m_flag = 1;
	for(int i = 0; i < 3; i++)
		for(int j = 0; j < NUMORD * NUMORD; j++)
			vec_lit[j][i] = vec_lit[j][m_flag];

	//to solve the linear equation to estimate lighting
	CvMat* pcoefmat = cvCreateMat(vec_normal.size(), NUMORD * NUMORD, CV_32FC1);
	CvMat* prgbmat  = cvCreateMat(vec_normal.size(), 1, CV_32FC1);
	CvMat* plitmat  = cvCreateMat(NUMORD * NUMORD, 1, CV_32FC1);

	for(unsigned int i = 0; i < vec_normal.size(); i++)
	{
		CVec3f m_nortmp = vec_normal[i];

		vector<float> vec_shtmp(NUMORD * NUMORD);
		GetSHval(m_nortmp, vec_shtmp);

		for (int j = 0; j < NUMORD * NUMORD; j++)
			cvmSet(pcoefmat, i, j, vec_shtmp[j]);
	}


	vector<CVec3f> vec_albedo(vec_pos.size());
	for(int nchal = 0; nchal < 3; nchal++)
	{
		for(int i = 0; i < NUMORD * NUMORD; i++)
			cvmSet(plitmat, i, 0, vec_lit[i][nchal]);

		cvMatMul(pcoefmat, plitmat, prgbmat);

		for(unsigned int i = 0; i < vec_albedo.size(); i++)
		{
			float tmpestal = (float)cvmGet(prgbmat, i, 0);
			vec_albedo[i][nchal] = vec_color[i][nchal] / tmpestal;
		}
	}

	//FILE* pnewfile = fopen("depthalbedo.txt", "w");
	//for(unsigned int i = 0; i < vec_pos.size(); i++)
	//	fprintf(pnewfile, "%d %d %f %f %f\n", vec_pos[i].x, vec_pos[i].y,
	//		vec_albedo[i][0], vec_albedo[i][1], vec_albedo[i][2]);
	//fclose(pnewfile);

	cvReleaseMat(&plitmat);
	cvReleaseMat(&pcoefmat);
	cvReleaseMat(&prgbmat);

	//see the albedo map
	IplImage* pimg = cvLoadImage(maskimgfile, 1);
	for(int i = 0; i < pimg->height; i++)
	{
		for(int j = 0; j < pimg->width; j++)
		{
			int nind = i * pimg->widthStep + 3 * j;

			pimg->imageData[nind]   = 127;
			pimg->imageData[nind+1] = 127;
			pimg->imageData[nind+2] = 127;
		}
	}

	for(unsigned int i = 0; i < vec_pos.size(); i++)
	{
		int nind = vec_pos[i].y * pimg->widthStep + 3 * vec_pos[i].x;

		pimg->imageData[nind]   = (BYTE)(clamp<float>(vec_albedo[i][2] * 128.0f, 0, 255));
		pimg->imageData[nind+1] = (BYTE)(clamp<float>(vec_albedo[i][1] * 128.0f, 0, 255));
		pimg->imageData[nind+2] = (BYTE)(clamp<float>(vec_albedo[i][0] * 128.0f, 0, 255));
	}

	cvSaveImage("albedo.png", pimg);
	cvReleaseImage(&pimg);

	printf("Done.\n");

	return 1;
}

int LightingEstimation(const char* maskimgfile, const vector<CvPoint>& vec_pos, const vector<CVec3f>& vec_normal, const vector<CVec3f>& vec_color)
{
	printf("Lighting Estimation ... ");

	//---- read mask image ------------------------------------------------------------------------
	IplImage* pmaskimg = cvLoadImage(maskimgfile, 0);

	int width  = pmaskimg->width;
	int height = pmaskimg->height;

	vector<vector<int>> vec_mask(height);

	for(int i = 0; i < height; i++)
		vec_mask[i].resize(width);

	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
		{
			int nind = i * pmaskimg->widthStep + j;
			BYTE m_val = pmaskimg->imageData[nind];

			if(m_val == 255) vec_mask[i][j] =  1;
			else             vec_mask[i][j] = -1;
		}
	}

	cvReleaseImage(&pmaskimg);

	//---- compute spherical harmonic coefficients for every pixel --------------------------------
	vector<vector<float>> mat_shval;
	vector<CVec3f> vec_rgb;
	for(unsigned int i = 0; i < vec_pos.size(); i++)
	{
		if (vec_mask[vec_pos[i].y][vec_pos[i].x] < 0)
			continue;

		vector<float> vec_sh(NUMORD * NUMORD);
		CVec3f m_nortmp = vec_normal[i];

		GetSHval(m_nortmp, vec_sh);

		mat_shval.push_back(vec_sh);
		vec_rgb.push_back(vec_color[i]);
	}

	//---- solve linear equation system to estimate lighting --------------------------------------
	CvMat* pcoefmat = cvCreateMat(mat_shval.size(), NUMORD * NUMORD, CV_32FC1);
	CvMat* prgbmat  = cvCreateMat(mat_shval.size(), 1, CV_32FC1);
	CvMat* plitmat  = cvCreateMat(NUMORD * NUMORD, 1, CV_32FC1);

	for(unsigned int i = 0; i < mat_shval.size(); i++)
		for (int j = 0; j < NUMORD * NUMORD; j++)
			cvmSet(pcoefmat, i, j, mat_shval[i][j]);

	FILE* plitfile = fopen("litcoef.txt", "w");

	for(int nchal = 0; nchal < 3; nchal++)
	{
		for (unsigned int i = 0; i < vec_rgb.size(); i++)
			cvmSet(prgbmat, i, 0, vec_rgb[i][nchal]);

		//solve the linear equation for illumination estimation
		cvSolve(pcoefmat, prgbmat, plitmat, CV_SVD);

		for(int i = 0; i < NUMORD * NUMORD; i++)
		{
			double tmpval = cvmGet(plitmat,i,0);
			fprintf(plitfile, "%f ", tmpval);
		}
		fprintf(plitfile, "\n");
	}
	fclose(plitfile);

	cvReleaseMat(&plitmat);
	cvReleaseMat(&pcoefmat);
	cvReleaseMat(&prgbmat);

	printf("Done.\n");
	return 1;
}

int main()
{
	g_scalefactor[0] =  (float)sqrt( 1 / PI) / 2;
	g_scalefactor[1] = -(float)sqrt( 3 / PI  / 4);
	g_scalefactor[2] =  (float)sqrt( 3 / PI  / 4);
	g_scalefactor[3] = -(float)sqrt( 3 / PI  / 4);
	g_scalefactor[4] =  (float)sqrt(15 / PI) / 2;
	g_scalefactor[5] = -(float)sqrt(15 / PI) / 2;
	g_scalefactor[6] =  (float)sqrt( 5 / PI) / 4;
	g_scalefactor[7] = -(float)sqrt(15 / PI) / 2;
	g_scalefactor[8] =  (float)sqrt(15 / PI) / 4;

	const char* depthfilename = "frame-hand-0726.txt";
	const char* maskfilename  = "mask-hand-fist.png";
	
	vector<CvPoint> pos;
	vector<CVec3f>  normal;
	vector<CVec3f>  color;

	double t = mtime();
	ReadFrame(depthfilename, pos, color, normal);
	printf("  time: %.3f s\n\n", mtime() - t); t = mtime();
	LightingEstimation(maskfilename, pos, color, normal);
	printf("  time: %.3f s\n\n", mtime() - t); t = mtime();
	AlbedoEstimation(  maskfilename, pos, color, normal);
	printf("  time: %.3f s\n\n", mtime() - t); t = mtime();

	return 1;
}