﻿// $Id: GaussianBlur.cs 528 2011-04-18 17:51:13Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using OpenTK.Graphics.OpenGL;

namespace VideoLib.OpenGL
{
	/// <summary>Applies a separable Gaussian blur filter.</summary>
	public class GaussianBlur : FboNode
	{
		/// <summary>Gets or sets the standard deviation of the blur (sigma).</summary>
		[Uniform(), Description("Standard deviation for filtering.")]
		public float Sigma { get; set; }

		/// <summary>Initialises a new instance of the <see cref="GaussianBlur"/> class.</summary>
		public GaussianBlur()
		{
			Sigma = 3;
			AddDefaultPins();
		}

		#region Rendering functions
//=================================================================================================

		private Texture intermediateTexture;

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// initialise intermediate texture
			intermediateTexture = new Texture(128, 128) { Wrap = TextureWrapMode.ClampToEdge };

			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.GaussianBlur.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Pin = "in", Uniform = "input", Unit = 0 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		/// 
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			return Draw(inputs["in"]);
		}

		private TextureFrame Draw(TextureFrame texture)
		{
			// resize intermediate texture if necessary (note: transposed relative to input texture)
			intermediateTexture.ReAlloc(texture.Height, texture.Width);

			// create output texture to fit input texture
			var resultTexture = TextureFrame.CreateToMatch(texture);
			resultTexture.Wrap = TextureWrapMode.ClampToEdge;
			resultTexture.Alloc();

			UpdateUniforms();

			// first pass of Gaussian blur
			texture.Bind(TextureUnit.Texture0);
			DrawScreenQuad(intermediateTexture);

			// second pass of Gaussian blur
			intermediateTexture.Bind(TextureUnit.Texture0);
			DrawScreenQuad(resultTexture);

			// cleanup
			intermediateTexture.Unbind();

			return resultTexture;
		}

//=================================================================================================
		#endregion Rendering functions
	}
}
