﻿// $Id: TextureInput.cs 526 2011-04-06 20:30:11Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;
using VideoLib.Frames;

namespace VideoLib.OpenGL
{
	/// <summary>Node that converts a <see cref="BitmapFrame"/> to a <see cref="TextureFrame"/>.</summary>
	public class TextureInput : Node, IOpenGLNode
	{
		#region Texture properties
//=================================================================================================

		/// <summary>Gets or sets the internal pixel format of the texture. Default: RGBA16F.</summary>
		public PixelInternalFormat InternalFormat { get; set; }

		/// <summary>Gets or sets the texture magnification filter of the texture. Default: linear.</summary>
		public TextureMagFilter MagFilter { get; set; }

		/// <summary>Gets or sets the texture minification filter of the texture. Default: linear.</summary>
		public TextureMinFilter MinFilter { get; set; }

		/// <summary>Sets both minification and magnification filters of the texture. Write-only.</summary>
		public TextureMagFilter Filter { set { MagFilter = value; MinFilter = (TextureMinFilter)(int)value; } }

		/// <summary>Gets or sets the horizontal texture wrap mode of the texture. Default: clamp.</summary>
		public TextureWrapMode WrapS { get; set; }

		/// <summary>Gets or sets the vertical texture wrap mode of the texture. Default: clamp.</summary>
		public TextureWrapMode WrapT { get; set; }

		/// <summary>Sets both texture wrap modes of the texture. Write-only.</summary>
		public TextureWrapMode Wrap { set { WrapS = value; WrapT = value; } }

//=================================================================================================
		#endregion // Texture properties

		/// <summary>Initialises a new instance of the <see cref="TextureInput"/> class.</summary>
		public TextureInput()
		{
			// default texture settings
			InternalFormat = PixelInternalFormat.Rgba16f;
			MagFilter = TextureMagFilter.Linear;
			MinFilter = TextureMinFilter.Linear;
			Wrap = TextureWrapMode.Clamp;

			AddInputPin("in");
			AddOutputPin("out");
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get the input bitmap data
			Frame frame = GetInputFrame("in");
			if (PassOnEndOfStreamFrame(frame)) return;
			BitmapFrame bitmap = CastFrameTo<BitmapFrame>(frame);

			// set up a new texture frame with the same size
			TextureFrame outputFrame = new TextureFrame(bitmap.Width, bitmap.Height)
			{
				InternalFormat = InternalFormat,
				MagFilter      = MagFilter,
				MinFilter      = MinFilter,
				WrapS          = WrapS,
				WrapT          = WrapT,
				SequenceNumber = bitmap.SequenceNumber
			};

			// copy in the image data
			outputFrame.Texture.Update(bitmap);

			// output the frame & clean up
			PushFrame(outputFrame);
			outputFrame.RefDec();
		}

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		/// <returns>A <see cref="System.String"/> that represents this instance.</returns>
		public override string ToString()
		{
			// make a list of items to mention (don't mention default settings)
			List<string> items = new List<string>();
			items.Add(InternalFormat.ToString());

			if ((int)MinFilter != (int)MagFilter) items.Add(String.Format("{0}/{1}", ToString(MinFilter), ToString(MagFilter)));
			else if(MinFilter != TextureMinFilter.Linear) items.Add(ToString(MinFilter));

			if ((int)WrapS != (int)WrapT) items.Add(String.Format("{0}/{1}", ToString(WrapS), ToString(WrapT)));
			else if (WrapS != TextureWrapMode.Clamp) items.Add(ToString(WrapS));

			return string.Format("{0}({1})", Name, string.Join(", ", items.ToArray()));
		}

		/// <summary>Returns a short <see cref="System.String"/> that represents a filter setting.</summary>
		/// <param name="filter">The minification filter.</param>
		private static string ToString(TextureMinFilter filter)
		{
			switch (filter)
			{
				case TextureMinFilter.Linear: return "Linear";
				case TextureMinFilter.LinearMipmapLinear: return "LinMipLin";
				case TextureMinFilter.LinearMipmapNearest: return "LinMipNN";
				case TextureMinFilter.Nearest: return "NN";
				case TextureMinFilter.NearestMipmapLinear: return "NNMipLin";
				case TextureMinFilter.NearestMipmapNearest: return "NNMipNN";
				default: return filter.ToString();
			}
		}

		/// <summary>Returns a short <see cref="System.String"/> that represents a filter setting.</summary>
		/// <param name="filter">The magnification filter.</param>
		private static string ToString(TextureMagFilter filter)
		{
			return ToString((TextureMinFilter)(int)filter);
		}

		/// <summary>Returns a short <see cref="System.String"/> that represents a texture wrap mode.</summary>
		/// <param name="wrap">The magnification filter.</param>
		private static string ToString(TextureWrapMode wrap)
		{
			switch (wrap)
			{
				case TextureWrapMode.ClampToBorder:  return "Border";
				case TextureWrapMode.ClampToEdge:    return "Edge";
				case TextureWrapMode.MirroredRepeat: return "Mirror";
				default:                             return wrap.ToString(); // Clamp, Repeat, and everything else
			}
		}
	}
}
