﻿// $Id: StereoSplitter.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.Drawing;

namespace VideoLib.Filters
{
	/// <summary>Implements a quick-and-dirty stereo video splitter.</summary>
	public class StereoSplitter : Node
	{
		/// <summary>Describes the layout of the left and right videos in the stereo video.</summary>
		[Flags]
		public enum VideoLayout
		{
			/// <summary>Auto-detect the layout, i.e. split along the longest dimension.</summary>
			Auto = 1,

			/// <summary>Videos are side-by-side.</summary>
			SideBySide = 2,

			/// <summary>Videos are on top of each other.</summary>
			OnTop = 4,

			/// <summary>The left-eye video is the first, i.e. left/top one.</summary>
			LeftFirst = 8,

			/// <summary>The right-eye video is the first, i.e. left/top one.</summary>
			RightFirst = 16,
		}

		/// <summary>Gets or sets the video layout.</summary>
		/// <value>The video layout.</value>
		public VideoLayout Layout { get; set; }

		/// <summary>Initializes a new instance of the <see cref="StereoSplitter"/> class.</summary>
		/// <param name="videoLayout">The layout of videos within the stereo video.</param>
		public StereoSplitter(VideoLayout videoLayout) : base()
		{
			Layout = videoLayout;
			AddInputPin("in");
			AddOutputPins("outL", "outR");
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="StereoSplitter"/> class.
		/// Auto-detects layout and assumes the first video is the right-eye video.
		/// </summary>
		public StereoSplitter()
			: this(VideoLayout.Auto | VideoLayout.RightFirst)
		{
		}

		/// <summary>Processes a frame.</summary>
		public override void Process()
		{
			// get input bitmap
			Frame frame = GetInputFrame("in");
			if (PassOnEndOfStreamFrame(frame)) return;
			BitmapFrame inputFrame = CastFrameTo<BitmapFrame>(frame);

			// auto-detect layout
			if ((Layout & VideoLayout.Auto) > 0)
			{
				if (inputFrame.Width > inputFrame.Height) Layout += VideoLayout.SideBySide - VideoLayout.Auto;
				else if (inputFrame.Width < inputFrame.Height) Layout += VideoLayout.OnTop - VideoLayout.Auto;
				else throw new Exception("Could not detect how to split video.");
			}

			// split the video frame according to config
			BitmapFrame outputFrame1, outputFrame2;
			switch (Layout & ~(VideoLayout.LeftFirst | VideoLayout.RightFirst))
			{
				case VideoLayout.SideBySide:
					SplitSideBySide(inputFrame, out outputFrame1, out outputFrame2); break;

				case VideoLayout.OnTop:
					SplitOnTop(inputFrame, out outputFrame1, out outputFrame2); break;

				default: throw new Exception("Unknown video Layout.");
			}

			// output the frames
			if ((Layout & VideoLayout.LeftFirst) > 0)
			{
				PushFrame("outL", outputFrame1);
				PushFrame("outR", outputFrame2);
			}
			else
			{
				PushFrame("outL", outputFrame2);
				PushFrame("outR", outputFrame1);
			}
		}

		/// <summary>Splits a stereo frame with "side-by-side" configuration.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outL">The left output frame.</param>
		/// <param name="outR">The right output frame.</param>
		private void SplitSideBySide(BitmapFrame inputFrame, out BitmapFrame outL, out BitmapFrame outR)
		{
			switch (inputFrame.Format)
			{
				case PixelFormat.BGR24: SplitSideBySideBGR24(inputFrame, out outL, out outR); break;
				case PixelFormat.BGR32: SplitSideBySideBGR32(inputFrame, out outL, out outR); break;

				default:
					throw new NotImplementedException(String.Format("Bitmap format \"{0}\" is not supported. Try BGR24 or BGR32.", inputFrame.Format));
			}
		}

		/// <summary>Splits an BGR24 stereo frame with "side-by-side" configuration.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outL">The left output frame.</param>
		/// <param name="outR">The right output frame.</param>
		private void SplitSideBySideBGR24(BitmapFrame inputFrame, out BitmapFrame outL, out BitmapFrame outR)
		{
			// create output frames
			outL = new BitmapFrame(inputFrame.Width / 2, inputFrame.Height, PixelFormat.BGR24) { SequenceNumber = inputFrame.SequenceNumber };
			outR = new BitmapFrame(inputFrame.Width / 2, inputFrame.Height, PixelFormat.BGR24) { SequenceNumber = inputFrame.SequenceNumber };

			// copy data around, be careful when inputFrame.Width is odd
			int halfRow = 3 * (inputFrame.Width / 2);
			for (int y = 0; y < inputFrame.Height; y++)
			{
				Array.Copy(inputFrame.Data, y * inputFrame.Stride, outL.Data, y * outL.Stride, halfRow);
				Array.Copy(inputFrame.Data, y * inputFrame.Stride + halfRow, outR.Data, y * outR.Stride, halfRow);
			}
		}

		/// <summary>Splits an BGR32 stereo frame with "side-by-side" configuration.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outL">The left output frame.</param>
		/// <param name="outR">The right output frame.</param>
		private void SplitSideBySideBGR32(BitmapFrame inputFrame, out BitmapFrame outL, out BitmapFrame outR)
		{
			// create output frames
			outL = new BitmapFrame(inputFrame.Width / 2, inputFrame.Height, PixelFormat.BGR32) { SequenceNumber = inputFrame.SequenceNumber };
			outR = new BitmapFrame(inputFrame.Width / 2, inputFrame.Height, PixelFormat.BGR32) { SequenceNumber = inputFrame.SequenceNumber };

			// copy data around, be careful when inputFrame.Width is odd
			int halfRow = 4 * (inputFrame.Width / 2);
			for (int y = 0; y < inputFrame.Height; y++)
			{
				Array.Copy(inputFrame.Data, y * inputFrame.Stride, outL.Data, y * outL.Stride, halfRow);
				Array.Copy(inputFrame.Data, y * inputFrame.Stride + halfRow, outR.Data, y * outR.Stride, halfRow);
			}
		}

		/// <summary>Splits a stereo frame with "on-top" configuration.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outL">The left output frame.</param>
		/// <param name="outR">The right output frame.</param>
		private void SplitOnTop(BitmapFrame inputFrame, out BitmapFrame outL, out BitmapFrame outR)
		{
			switch (inputFrame.Format)
			{
				case PixelFormat.BGR24: SplitOnTopBGR24(inputFrame, out outL, out outR); break;
				case PixelFormat.BGR32: SplitOnTopBGR32(inputFrame, out outL, out outR); break;

				default:
					throw new NotImplementedException(String.Format("Bitmap format \"{0}\" is not supported. Try BGR24 or BGR32.", inputFrame.Format));
			}
		}

		/// <summary>Splits an BGR24 stereo frame with "on-top" configuration.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outL">The left output frame.</param>
		/// <param name="outR">The right output frame.</param>
		private void SplitOnTopBGR24(BitmapFrame inputFrame, out BitmapFrame outL, out BitmapFrame outR)
		{
			// create output frames
			outL = new BitmapFrame(inputFrame.Width, inputFrame.Stride, inputFrame.Height / 2,
				PixelFormat.BGR24) { SequenceNumber = inputFrame.SequenceNumber };
			outR = new BitmapFrame(inputFrame.Width, inputFrame.Stride, inputFrame.Height / 2,
				PixelFormat.BGR24) { SequenceNumber = inputFrame.SequenceNumber };

			// copy data around
			Array.Copy(inputFrame.Data, 0, outL.Data, 0, outL.Height * outL.Stride);
			Array.Copy(inputFrame.Data, outL.Height * outL.Stride, outR.Data, 0, outL.Height * outL.Stride);
		}

		/// <summary>Splits an BGR32 stereo frame with "on-top" configuration.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outL">The left output frame.</param>
		/// <param name="outR">The right output frame.</param>
		private void SplitOnTopBGR32(BitmapFrame inputFrame, out BitmapFrame outL, out BitmapFrame outR)
		{
			// create output frames
			outL = new BitmapFrame(inputFrame.Width, inputFrame.Stride, inputFrame.Height / 2,
				PixelFormat.BGR32) { SequenceNumber = inputFrame.SequenceNumber };
			outR = new BitmapFrame(inputFrame.Width, inputFrame.Stride, inputFrame.Height / 2,
				PixelFormat.BGR32) { SequenceNumber = inputFrame.SequenceNumber };

			// copy data around
			Array.Copy(inputFrame.Data, 0, outL.Data, 0, outL.Height * outL.Stride);
			Array.Copy(inputFrame.Data, outL.Height * outL.Stride, outR.Data, 0, outL.Height * outL.Stride);
		}
	}
}
