﻿// $Id: SwitchFilter.cs 521 2011-01-04 17:39:18Z cr333 $
using System.Collections.Generic;
using System.Linq;
using VideoLib.Frames;

namespace VideoLib.Filters
{
	/// <summary>Connects incoming frames to several output nodes, forwarding to one (or more) selected node(s) only.</summary>
	public class SwitchFilter : Node
	{
		/// <summary>Gets or sets the number of outputs to support.</summary>
		public int OutputCount { get; protected set; }

		/// <summary>Gets or sets the pin names.</summary>
		public string[] PinNames { get; protected set; }

		/// <summary>Gets or sets the selected output(s).</summary>
		public int SelectedOutput { get; set; }

		/// <summary>Gets or sets whether multiple outputs are supported. Interprets <see cref="SelectedOutput"/> as binary number. Default: false.</summary>
		public bool MultipleOutputs { get; set; }

		/// <summary>Initialises a new instance of the <see cref="SwitchFilter"/> class.</summary>
		/// <param name="outputCount">The number of outputs.</param>
		/// <param name="pinNames">The pin names.</param>
		public SwitchFilter(int outputCount, params string[] pinNames)
		{
			PinNames        = pinNames;
			OutputCount     = outputCount;
			SelectedOutput  = 0;
			MultipleOutputs = false;

			// create the IO pins
			foreach (var pinName in PinNames)
			{
				AddInputPin(pinName + "In");
				for (int i = 0; i < OutputCount; i++)
					AddOutputPin(string.Format("{0}Out{1}", pinName, i));
			}
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get latest frames
			var frames = GetInputFrames(InputPins.Keys.ToArray());
			if (PassOnEndOfStreamFrame(frames)) return;

			foreach (var pinName in PinNames)
			{
				var frame = frames[pinName + "In"];

				// 'SelectedOutput' is passed to avoid changes of its value while pushing out frames
				foreach(var output in EnumerateOutputs(SelectedOutput))
					PushFrame(string.Format("{0}Out{1}", pinName, output), frame);

				// decrement reference counter if required
				var refCounted = frame as IRefCountedFrame;
				if (refCounted != null) refCounted.RefDec();
			}
		}

		/// <summary>Enumerates all selected outputs.</summary>
		/// <param name="output">Selected outputs indicator.</param>
		protected IEnumerable<int> EnumerateOutputs(int output)
		{
			if (MultipleOutputs)
			{
				// interpret 'output' in binary fashion
				for (int i = 0; i < OutputCount; i++)
					if ((output & (1 << i)) > 0)
						yield return i;
			}
			else
			{
				// just check for valid range
				if (output >= 0 && output < OutputCount)
					yield return SelectedOutput;
			}
		}
	}
}
