﻿// $Id: BitmapFrame.cs 370 2010-08-30 14:08:55Z cr333 $
using System;
using System.Drawing;
using System.Runtime.InteropServices;
using VideoLib.Filters;
using VideoLib.Utils;
using Imaging = System.Drawing.Imaging;

namespace VideoLib.Frames
{
	/// <summary>A video frame based on a byte buffer.</summary>
	public class BitmapFrame : VideoFrame, IDisposable
	{
		#region Properties

		// NB: Width & Height are defined in superclass VideoFrame.

		/// <summary>
		/// Gets the stride of the bitmap frame.
		/// </summary>
		public int Stride { get; protected set; }

		/// <summary>
		/// Gets the data buffer of the bitmap frame.
		/// </summary>
		public byte[] Data { get; protected set; }

		/// <summary>
		/// Memory allocator that is used for allocating byte arrays.
		/// Default: memory pool.
		/// </summary>
		public static MemoryAllocator Allocator { get; set; }

		/// <summary>
		/// The allocator used for a particular instance of BitmapFrame.
		/// </summary>
		/// <remarks>
		/// This is stored for each instance, as the static allocator (above) might change before
		/// this frame is freed -- and we don't want to free it with the wrong allocator, do we?
		/// </remarks>
		private MemoryAllocator AllocatorUsed { get; set; }

		#endregion // Properties

		#region Constructors

		/// <summary>Static constructor that sets the default memory allocator.</summary>
		static BitmapFrame()
		{
			Allocator = new BuiltinAllocator(); // using new & GC
			//Allocator = new MultiMemoryPoolAllocator(); // not tested thoroughly; occasionally causes OutOfMemoryException
		}

		/// <summary>Creates an empty bitmap frame.</summary>
		/// <param name="width">The width of the bitmap.</param>
		/// <param name="height">The height of the bitmap.</param>
		/// <param name="format">The format of the bitmap.</param>
		public BitmapFrame(int width, int height, PixelFormat format)
			: base(width, height, format)
		{
			// first compute the minimum stride
			int minStride = width * format.BytesPerPixel();

			// then round up to nearest word (4 bytes)
			Stride = (minStride + 3) & ~3;

			AllocatorUsed = Allocator;
			Data = AllocatorUsed.AllocArray(Stride * Height);
		}

		/// <summary>Creates an empty bitmap frame.</summary>
		/// <param name="width">The width of the bitmap.</param>
		/// <param name="stride">The stride (in bytes) of a row in the bitmap.</param>
		/// <param name="height">The height of the bitmap.</param>
		/// <param name="format">The format of the bitmap.</param>
		public BitmapFrame(int width, int stride, int height, PixelFormat format)
			: base(width, height, format)
		{
			// first compute the minimum stride
			int minStride = width * format.BytesPerPixel();

			// then round up to nearest word (4 bytes)
			minStride = (minStride + 3) & ~3;

			if (stride < minStride)
				throw new ArgumentException(String.Format("Invalid stride (stride={0}, width={1}, format={2}). Must be at least {3}.", stride, width, format, minStride), "stride");

			Stride = stride;
			AllocatorUsed = Allocator;
			Data = AllocatorUsed.AllocArray(Stride * Height);
		}

		/// <summary>Creates a bitmap frame by copying data into it.</summary>
		/// <param name="scan0">The pointer to the first element in the bitmap.</param>
		/// <param name="width">The width of the bitmap.</param>
		/// <param name="stride">The stride (in bytes) of a row in the bitmap.</param>
		/// <param name="height">The height of the bitmap.</param>
		/// <param name="format">The format of the bitmap.</param>
		public BitmapFrame(IntPtr scan0, int width, int stride, int height, PixelFormat format)
			: this(width, stride, height, format)
		{
			Marshal.Copy(scan0, Data, 0, Stride * Height);
		}
		
		/// <summary>Creates a new frame with the same size, format and sequence number.</summary>
		/// <param name="other">The template frame to base the new frame on.</param>
		public static BitmapFrame CreateToMatch(BitmapFrame other)
		{
			return new BitmapFrame(other.Width, other.Stride, other.Height, other.Format) { SequenceNumber = other.SequenceNumber };
		}
		
		#endregion // Constructors

		#region Cleanup

		/// <summary>
		/// Frees the associated bitmap data.
		/// </summary>
		public void Dispose()
		{
			AllocatorUsed.FreeArray(Data);
			Data = null;
		}

		/// <summary>
		/// Cleans up the BitmapFrame by freeing its data.
		/// </summary>
		~BitmapFrame()
		{
			AllocatorUsed.FreeArray(Data);
			Data = null;
		}

		#endregion // Cleanup

		#region From/to Drawing.Bitmap

		/// <summary>
		/// Initializes a new instance of the <see cref="BitmapFrame"/> class from a given Bitmap.
		/// </summary>
		/// <param name="bitmap">The bitmap.</param>
		public BitmapFrame(System.Drawing.Bitmap bitmap)
			: base()
		{
			FromBitmap(bitmap);
		}

		/// <summary>Copy image data from a given Drawing.Bitmap.</summary>
		/// <param name="bitmap">Bitmap to use.</param>
		protected virtual void FromBitmap(System.Drawing.Bitmap bitmap)
		{
			// make sure that we can read the bitmap, if not convert it
			Bitmap source = bitmap;
			if (bitmap.VideoLibPixelFormat() == PixelFormat.Unknown)
			{
				// unknown format => convert to BGR24
				source = new Bitmap(bitmap.Width, bitmap.Height, Imaging.PixelFormat.Format24bppRgb);
				using (Graphics g = Graphics.FromImage(source))
					g.DrawImageUnscaled(bitmap, 0, 0);
			}

			// copy the bitmap's size and format
			Width = source.Width;
			Height = source.Height;
			Format = source.VideoLibPixelFormat();

			// lock the bitmap's bits
			Imaging.BitmapData bitmapData = new Imaging.BitmapData();
			source.LockBits(new Rectangle(0, 0, source.Width, source.Height),
				Imaging.ImageLockMode.ReadOnly, source.PixelFormat, bitmapData);

			// copy out the bitmap data
			Stride = bitmapData.Stride;
			AllocatorUsed = Allocator;
			Data = AllocatorUsed.AllocArray(Stride * Height);
			Marshal.Copy(bitmapData.Scan0, Data, 0, Stride * Height);
			source.UnlockBits(bitmapData);

			// free temporary bitmap
			if (bitmap.VideoLibPixelFormat() == PixelFormat.Unknown)
				source.Dispose();
		}

		/// <summary>Copies the frame's data to a <see cref="Bitmap"/>.</summary>
		/// <param name="bitmap">Bitmap to copy frame contents to. Use null Bitmap to create a new one.</param>
		public virtual void FillBitmap(ref System.Drawing.Bitmap bitmap)
		{
			// first check if the pixel format is supported
			var bitmapFrame   = this;
			var imagingFormat = bitmapFrame.Format.ToImagingPixelFormat();
			if (imagingFormat == Imaging.PixelFormat.Undefined)
			{
				// try converting to byte-based bitmap
				bitmapFrame   = ConvertFilter.ConvertFormat(bitmapFrame, (bitmapFrame.Format & ~PixelFormat.TypeMask) | PixelFormat.Byte);
				imagingFormat = bitmapFrame.Format.ToImagingPixelFormat();
			}

			// then check again if format is supported
			if (imagingFormat == Imaging.PixelFormat.Undefined)
				throw new NotSupportedException(String.Format("Pixel format {0} cannot be converted to a Bitmap.", Format));

			// check for valid Bitmap, size and format
			if (bitmap == null || bitmap.Width != Width || bitmap.Height != Height || bitmap.PixelFormat != imagingFormat)
				bitmap = new Bitmap(Width, Height, imagingFormat);

			// lock bits for writing and copy in the data
			Imaging.BitmapData bitmapData = new Imaging.BitmapData();
			bitmap.LockBits(
				new Rectangle(0, 0, bitmap.Width, bitmap.Height),
				Imaging.ImageLockMode.WriteOnly, bitmap.PixelFormat, bitmapData);

			// if both we and the bitmap use the same stride, we can copy one large array
			if (bitmapFrame.Stride == bitmapData.Stride)
				Marshal.Copy(bitmapFrame.Data, 0, bitmapData.Scan0, bitmapData.Stride * bitmap.Height);
			else
			{
				// copy bitmap between buffers using different strides
				for (int y = 0; y < bitmap.Height; y++)
				{
					Marshal.Copy(
						bitmapFrame.Data,
						y * bitmapFrame.Stride,
						new IntPtr(bitmapData.Scan0.ToInt64() + y * bitmapData.Stride),
						Math.Min(Stride, bitmapData.Stride)
					);
				}

			}

			bitmap.UnlockBits(bitmapData);

			// if this is an 8/bit greyscale image, make sure the palette is set correctly
			if (bitmap.PixelFormat == Imaging.PixelFormat.Format8bppIndexed)
			{
				Imaging.ColorPalette palette = bitmap.Palette;
				for (int x = 0; x < 256; x++) palette.Entries[x] = Color.FromArgb(x, x, x);
				bitmap.Palette = palette;
			}
		}

		/// <summary>
		/// Convenience access to the underlying bitmap.
		/// </summary>
		public System.Drawing.Bitmap Bitmap
		{
			get
			{
				System.Drawing.Bitmap b = null;
				FillBitmap(ref b);
				return b;
			}
			protected set
			{
				FromBitmap(value);
			}
		}

		#endregion

		#region Pixel format conversion

		/// <summary>
		/// Gets the pixel format as Imaging.PixelFormat.
		/// </summary>
		protected System.Drawing.Imaging.PixelFormat ImagingPixelFormat { get { return Format.ToImagingPixelFormat(); } }

		/// <summary>
		/// Gets the pixel format as Imaging.PixelFormat, mapping Undefined to BGR24.
		/// </summary>
		[Obsolete]
		protected System.Drawing.Imaging.PixelFormat PixelFormatOrBGR24
		{
			get
			{
				if (ImagingPixelFormat == System.Drawing.Imaging.PixelFormat.Undefined)
					return System.Drawing.Imaging.PixelFormat.Format24bppRgb;
				else
					return ImagingPixelFormat;
			}
		}

		#endregion

		/// <summary>
		/// Returns a <see cref="System.String"/> that represents this instance.
		/// </summary>
		public override string ToString()
		{
			return String.Format("BitmapFrame #{0} ({1}, {2}x{3})",
				SequenceNumber, Format.ToString(), Width, Height);
		}

		/// <summary>
		/// Copies a frame.
		/// </summary>
		/// <param name="newFrame">The destination Frame where this Frame's data should be copied to.</param>
		public override sealed void copyTo(ref Frame newFrame)
		{
			if (newFrame == null)
				newFrame = new BitmapFrame(Width, Stride, Height, Format);

			if (newFrame.GetType() != this.GetType() || ((BitmapFrame)newFrame).Format != Format)
				throw new ArgumentException("Cannot duplicate to a frame of a different type.");

			Array.Copy(Data, ((BitmapFrame)newFrame).Data, Data.Length);
			newFrame.SequenceNumber = SequenceNumber;
		}
	}
}
