﻿// $Id: INode.cs 549 2012-01-17 18:12:19Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.Diagnostics;
using System.Threading;

namespace VideoLib
{
	/// <summary>Possible states of a node.</summary>
	public enum NodeStatus
	{
		/// <summary>Unknown status. This is the default just after startup.</summary>
		Unknown,

		/// <summary>The node is not ready to run, because it is waiting for input.</summary>
		Waiting,

		/// <summary>The node is ready to process data.</summary>
		Ready,

		/// <summary>The node is currently running on a worker thread.</summary>
		Running,

		/// <summary>The node has finished processing frames.</summary>
		Finished,

		/// <summary>The node has stopped due to an error or user decision (e.g. closing a window).</summary>
		Stopped,
	}

	/// <summary>Event arguments for a change in node status.</summary>
	[DebuggerDisplay("{OldStatus} => {NewStatus}")]
	public class NodeStatusChangedEventArgs : EventArgs
	{
		/// <summary>Gets the old status.</summary>
		public NodeStatus OldStatus { get; protected set; }

		/// <summary>Gets the new status.</summary>
		public NodeStatus NewStatus { get; protected set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="NodeStatusChangedEventArgs"/> class.
		/// </summary>
		/// <param name="oldStatus">The old status.</param>
		/// <param name="newStatus">The new status.</param>
		public NodeStatusChangedEventArgs(NodeStatus oldStatus, NodeStatus newStatus)
		{
			OldStatus = oldStatus;
			NewStatus = newStatus;
		}
	}

	/// <summary>Handler for a node status change.</summary>
	public delegate void NodeStatusChangedHandler(object sender, NodeStatusChangedEventArgs e);

	/// <summary>Basic interface of a processing node.</summary>
	public interface INode : IDisposable
	{
		/// <summary>Name of a node.</summary>
		string Name { get; }

		/// <summary>The current status of a node.</summary>
		NodeStatus Status { get; }

		/// <summary>Parent topology.</summary>
		Topology Topology { get; }

		/// <summary>Input pins of a node.</summary>
		Dictionary<string, InputPin> InputPins { get; }

		/// <summary>Output pins of a node.</summary>
		Dictionary<string, OutputPin> OutputPins { get; }

		/// <summary>Notifies listeners about status changes.</summary>
		event NodeStatusChangedHandler StatusChanged;

		/// <summary>Reacts to new Frames arriving at the InputPins, but don't actually run this node yet.</summary>
		/// <param name="sender">The sender of the event.</param>
		/// <param name="e">New incoming Frame.</param>
		void NewFrameListener(object sender, NewFrameEventArgs e);

		/// <summary>Initialisation code, executed just before the first StatusUpdate.</summary>
		void Init();

		/// <summary>Anything that needs doing before Process() is called.</summary>
		void PreProcess();

		/// <summary>Processes some frames, to produce new frames.</summary>
		void Process();

		/// <summary>Anything that needs doing after Process() is called.</summary>
		void PostProcess();

		/// <summary>Cleanup code, executed when stopping the scheduler.</summary>
		void Stop();

		///// <summary>
		///// Set some algorithm parameters.
		///// <para>These are coded as strings to provide a uniform interface even when parameters
		///// have to be sent to PTs located on other computers in the network.</para>
		///// </summary>
		///// <param name="name">Parameter name to set.</param>
		///// <param name="value">New parameter value.</param>
		///// <returns>True iff setting the parameter succeeded.</returns>
		//bool setParameter(string name, string value);

		/// <summary>Request the node to update its status.</summary>
		void UpdateStatus();
	}

	/// <summary>
	/// Marker interface to indicate that this node contains CUDA code. 
	/// </summary>
	public interface ICudaNode : IDisposable
	{
	}

}
