﻿// $Id: Node.cs 549 2012-01-17 18:12:19Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.Threading;
using System.ComponentModel;

namespace VideoLib
{
	/// <summary>Abstract base class for processing nodes.</summary>
	/// <remarks>Implements most common functionality.</remarks>
	[System.Diagnostics.DebuggerDisplay("{Name} ({Status})")]
	public abstract class Node : INode, IDisposable
	{
		/// <summary>Name of this node.</summary>
		[Category("General"), ReadOnly(true), Description("The name of this node.")]
		public virtual string Name { get; protected set; }

		/// <summary>Indicates if this <see cref="Node"/> produces verbose output.</summary>
		[Category("General"), DefaultValue(false), Description("Indicates of this node produces verbose command line output.")]
		public bool Verbose { get; set; }

		/// <summary>Constructs a new abstract node, linked to the default topology.</summary>
		/// <remarks>To link a node to a different topology, just assign to Node.Topology.</remarks>
		public Node()
		{
			// get name of this node from the node type
			Name = GetType().Name;
			Verbose = false;

			// start with no input/output pins
			InputPins = new Dictionary<string, InputPin>();
			OutputPins = new Dictionary<string, OutputPin>();

			// link into the default topology (can change that later)
			Topology = Topology.MainInstance;
		}

		/// <summary>Returns the node name with some more details about the IO pins.</summary>
		public string PrintDetail()
		{
			return Name + " with (" + String.Join(", ", InputPins.Keys.ToArray()) + ") => (" + String.Join(", ", OutputPins.Keys.ToArray()) + ")";
		}

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		public override string ToString()
		{
			return Name;
		}

		#region IDisposable implementation
		//=========================================================================================

		// See <http://msdn.microsoft.com/en-us/library/system.gc.suppressfinalize.aspx> and 
		// <http://www.bluebytesoftware.com/blog/PermaLink.aspx?guid=88e62cdf-5919-4ac7-bc33-20c06ae539ae>
		// for articles on how IDisposable is meant to be implemented.
		// Also see <http://www.codeproject.com/KB/mcpp/CppDerivingDispose.aspx> for implementing the derived class in C++/CLI.

		/// <summary>Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.</summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		/// <summary>
		/// Releases unmanaged resources and performs other cleanup operations before the
		/// <see cref="Node"/> is reclaimed by garbage collection.
		/// </summary>
		~Node()
		{
			Dispose(false);
		}

		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (disposing)
			{
				// free managed resources
			}
	
			// free native resource

			// derived classes: base.Dispose(disposing)
		}

		//=========================================================================================
		#endregion // IDisposable implementation

		#region Main public interface
		//=========================================================================================

		/// <summary>Initialisation code, executed just before the first StatusUpdate.</summary>
		public virtual void Init()
		{
		}

		/// <summary>Sets the node status to Running before processing frames.</summary>
		public void PreProcess()
		{
			lock (this)
			{
				if (Status != NodeStatus.Stopped)
					Status = NodeStatus.Running;
			}
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public abstract void Process();

		/// <summary>Updates the node's status after processing frames finished.</summary>
		public void PostProcess()
		{
			lock (this)
			{
				if (Status != NodeStatus.Finished && Status != NodeStatus.Ready)
					Status = NodeStatus.Waiting;
				UpdateStatus();
			}
		}

		/// <summary>Cleanup code, executed when stopping the scheduler.</summary>
		public virtual void Stop()
		{
			// Tell the topology that this node has finished.
			Topology.NodeFinished();
		}

		//=========================================================================================
		#endregion Main public interface

		#region Status handling code
		//=========================================================================================

		/// <summary>The current status of a node.</summary>
		[BrowsableAttribute(false)]
		public virtual NodeStatus Status
		{
			get { return status; }
			protected set
			{
				if (status != value)
				{
					NodeStatus oldStatus = status;
					status = value;
					OnStatusChanged(oldStatus, status);
				}
			}
		}

		private NodeStatus status = NodeStatus.Unknown;
		//private Object statusLock = new Object();

		/// <summary>Notifies listeners about status changes.</summary>
		public event NodeStatusChangedHandler StatusChanged;

		/// <summary>Invokes the StatusChanged event.</summary>
		/// <param name="oldStatus">Previous status of the node.</param>
		/// <param name="newStatus">New status of the node.</param>
		protected virtual void OnStatusChanged(NodeStatus oldStatus, NodeStatus newStatus)
		{
			//Console.WriteLine(Name + ": " + oldStatus.ToString() + " => " + newStatus.ToString());
			if (StatusChanged != null)
				StatusChanged(this, new NodeStatusChangedEventArgs(oldStatus, newStatus));

			if (status == NodeStatus.Finished)
				Topology.NodeFinished();
		}

		/// <summary>Reacts to new Frames arriving at the InputPins, but don't actually run this node yet.</summary>
		/// <param name="sender">The sender of the event.</param>
		/// <param name="e">New incoming Frame.</param>
		public virtual void NewFrameListener(object sender, NewFrameEventArgs e)
		{
			UpdateStatus();
		}

		private Object statusLock = new Object();

		/// <summary>Request the node to update its status using UpdateStatusImpl.</summary>
		public void UpdateStatus() { lock (this) { UpdateStatusImpl(); } }

		/// <summary>Request the node to update its status.</summary>
		protected virtual void UpdateStatusImpl()
		{
			// Finished is a final status, cannot change to anything else later
			if (Status == NodeStatus.Finished) return;

			// Stopped is a final status, cannot change to anything else later (for now)
			if (Status == NodeStatus.Stopped) return;

			// don't update status if node is currently Running, to avoid multiple threads running it simultanously
			if (Status == NodeStatus.Running) return;

			// don't update status if node is ready to run
			if (Status == NodeStatus.Ready) return;

			// check if all input pins have data
			foreach (InputPin inputPin in InputPins.Values)
			{
				if (inputPin.DataAvailable == false)
				{
					Status = NodeStatus.Waiting;
					return;
				}
			}

			// check if all output pins are ready
			foreach (OutputPin outputPin in OutputPins.Values)
			{
				if (outputPin.Status != PinStatus.Ready)
				{
					Status = NodeStatus.Waiting;
					return;
				}
			}

			// if control reaches this, we've got all the data we need, so ready to run
			Status = NodeStatus.Ready;
		}

		//=========================================================================================
		#endregion Status handling code

		#region Topology handling code
		//=========================================================================================

		/// <summary>Parent topology.</summary>
		[BrowsableAttribute(false)]
		public Topology Topology
		{
			get { return topology; }
			set
			{
				// remove from existing topology first
				if (this.topology != null)
					this.topology.RemoveNode(this);

				// link into the given topology
				this.topology = value;
				if (this.topology != null)
					this.topology.AddNode(this);
			}
		}

		private Topology topology = null;

		//=========================================================================================
		#endregion Topology handling code

		#region IO pin handling code
		//=========================================================================================

		/// <summary>Input pins of this node.</summary>
		[BrowsableAttribute(false)]
		public Dictionary<string, InputPin> InputPins { get; private set; }

		/// <summary>Output pins of this node.</summary>
		[BrowsableAttribute(false)]
		public Dictionary<string, OutputPin> OutputPins { get; private set; }

		/// <summary>Finds the pin of a given name.</summary>
		/// <remarks>First looks for input, then output pins.</remarks>
		/// <param name="pinName">Name of the pin to find.</param>
		/// <returns>The found pin.</returns>
		public Pin this[string pinName]
		{
			get
			{
				if (InputPins.ContainsKey(pinName))  return InputPins[pinName];
				if (OutputPins.ContainsKey(pinName)) return OutputPins[pinName];
				throw new PinNotFoundException(String.Format("Could not find pin \"{0}\".", pinName));
			}
		}

		/// <summary>Adds a new input pin of the given name.</summary>
		/// <param name="name">The name of the new input pin.</param>
		protected void AddInputPin(string name)
		{
			if (InputPins.Keys.Contains(name))
				throw new Exception(String.Format("Input pin with name \"{0}\" already exists.", name));

			InputPins.Add(name, new InputPin(name, this));
		}

		/// <summary>Adds multiple input pins at the same time.</summary>
		/// <param name="names">Names of input pins.</param>
		protected void AddInputPins(params string[] names)
		{
			foreach (string name in names)
				AddInputPin(name);
		}

		/// <summary>Adds a new output pin of the given name.</summary>
		/// <param name="name">The name of the new output pin.</param>
		protected void AddOutputPin(string name)
		{
			if (OutputPins.Keys.Contains(name))
				throw new Exception(String.Format("Output pin with name \"{0}\" already exists.", name));

			OutputPins.Add(name, new OutputPin(name, this));
		}

		/// <summary>Adds multiple output pins at the same time.</summary>
		/// <param name="names">Names of output pins.</param>
		protected void AddOutputPins(params string[] names)
		{
			foreach (string name in names)
				AddOutputPin(name);
		}

		/// <summary>Adds the default IO pins ("in" and "out").</summary>
		protected void AddDefaultPins()
		{
			AddInputPin("in");
			AddOutputPin("out");
		}

		/// <summary>Gets an input pin by name.</summary>
		/// <param name="name">The name of the input pin.</param>
		/// <returns>The InputPin of the given name; null if not found.</returns>
		public InputPin GetInputPin(string name)
		{
			if (!InputPins.ContainsKey(name))
			{
				return null;
			}
			else
			{
				return InputPins[name];
			}
		}

		/// <summary>Returns the default input pin "in".</summary>
		/// <returns>Default input pin.</returns>
		public InputPin GetInputPin()
		{
			return GetInputPin("in");
		}

		/// <summary>Gets an output pin by name.</summary>
		/// <param name="name">The name of the output pin.</param>
		/// <returns>The output pin of the given name; throws ArgumentException if not found.</returns>
		public OutputPin GetOutputPin(string name)
		{
			if (!OutputPins.ContainsKey(name))
				throw new ArgumentException(String.Format("Output pin {0}[{1}] does not exist.", Name, name), "name");

			return OutputPins[name];
		}

		/// <summary>Returns the default output pin "out".</summary>
		/// <returns>Default output pin.</returns>
		public OutputPin GetOutputPin()
		{
			return GetOutputPin("out");
		}

		//=========================================================================================
		#endregion IO pin handling code

		#region Frame handling code
		//=========================================================================================

		/// <summary>Gets the input frame from the default input pin "in".</summary>
		/// <returns>The returned frame.</returns>
		public Frame GetInputFrame()
		{
			return GetInputFrame("in");
		}

		/// <summary>Gets an input frame from some input pin.</summary>
		/// <param name="inputPinName">Name of input pin to get frame from.</param>
		/// <returns>The returned frame.</returns>
		public Frame GetInputFrame(string inputPinName)
		{
			return GetInputPin(inputPinName).GetFrame();
		}

		/// <summary>Gets an input frame of some given type.</summary>
		/// <typeparam name="T">Desired type of frame, e.g. BitmapFrame.</typeparam>
		/// <param name="inputPinName">Name of input pin to get frame from.</param>
		/// <returns>Frame of the desired type, otherwise FrameNotSupportedException is thrown.</returns>
		public T GetInputFrame<T>(string inputPinName) where T : Frame
		{
			Frame frame = GetInputFrame(inputPinName);
			return CastFrameTo<T>(frame);
		}

		/// <summary>Gets a set of input frames by name of the input pin.</summary>
		/// <param name="inputPinNames">Names of the input pins to get frames from.</param>
		/// <returns>The returned frames, as a Dictionary indexed by pin name.</returns>
		public Dictionary<string, Frame> GetInputFrames(params string[] inputPinNames)
		{
			var frames = new Dictionary<string, Frame>();
			foreach (string inputPinName in inputPinNames)
			{
				frames.Add(inputPinName, GetInputFrame(inputPinName));
			}
			return frames;
		}

		/// <summary>Gets all input frames.</summary>
		/// <returns>The returned frames, as a Dictionary indexed by pin name.</returns>
		public Dictionary<string, Frame> GetInputFrames()
		{
			return GetInputFrames(InputPins.Keys.ToArray());
		}

		/// <summary>Converts a frame to some given type.</summary>
		/// <typeparam name="T">Desired type of frame, e.g. BitmapFrame.</typeparam>
		/// <param name="frame">The frame to convert.</param>
		/// <returns>Frame of the desired type, otherwise FrameNotSupportedException is thrown.</returns>
		public T CastFrameTo<T>(Frame frame) where T : Frame
		{
			if (!(frame is T)) throw new FrameNotSupportedException(frame);
			return (T)frame;
		}

		/// <summary>Converts a dictionary of frames to some given type.</summary>
		/// <typeparam name="T">Desired type of the frames, e.g. BitmapFrame.</typeparam>
		/// <param name="frames">A dictionary of Frames, indexed by pin name.</param>
		/// <returns>A dictionary of frames of the desired type,indexed by pin name; otherwise FrameNotSupportedException is thrown.</returns>
		public Dictionary<string, T> CastFramesTo<T>(Dictionary<string, Frame> frames)
			where T : Frame
		{
			Dictionary<string, T> result = new Dictionary<string, T>();

			foreach (var frame in frames)
			{
				if (frame.Value is T) result.Add(frame.Key, (T)frame.Value);
				else throw new FrameNotSupportedException(frame.Value);
			}

			return result;
		}

		/// <summary>Pushes an output frame through an output pin.</summary>
		/// <remarks>outputFrame is of type Object to allow PushFrame(..., frame.Clone()) without extra casts.</remarks>
		/// <param name="outputPinName">Name of the output pin to push frame through.</param>
		/// <param name="outputFrame">The frame to push.</param>
		public void PushFrame(string outputPinName, Object outputFrame)
		{
			if (outputFrame is Frame)
				GetOutputPin(outputPinName).PushFrame((Frame)outputFrame);
			else
				throw new ArgumentException("Invalid output frame.", "outputFrame");
		}

		/// <summary>Pushes an output frame through the default output pin "out".</summary>
		/// <remarks>outputFrame is of type Object to allow PushFrame(frame.Clone()) without extra casts.</remarks>
		/// <param name="outputFrame">The frame to push.</param>
		public void PushFrame(Object outputFrame)
		{
			PushFrame("out", outputFrame);
		}

		/// <summary>Checks if the given frame is an EOS one, and if so passes it on.</summary>
		/// <param name="frame">Frame to check for EOS-ness.</param>
		/// <returns>True if and only if an EOS frame has been detected and passed on.</returns>
		public virtual bool PassOnEndOfStreamFrame(Frame frame)
		{
			if (frame is EndOfStreamFrame)
			{
				foreach(var outputPin in OutputPins.Values)
					outputPin.PushFrame(frame);
				Status = NodeStatus.Finished;
				return true;
			}
			else
			{
				return false;
			}
		}

		/// <summary>Checks if any of the given frames is an EOS one, and if so passes it on to all output pins.</summary>
		/// <param name="frames">Frames to check for EOS-ness.</param>
		/// <returns>True if and only if an EOS frame has been detected and passed on.</returns>
		public bool PassOnEndOfStreamFrame(Dictionary<string, Frame> frames)
		{
			foreach (var frame in frames)
			{
				if (PassOnEndOfStreamFrame(frame.Value))
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Enumeration of (Bitmap)Frame properties.
		/// </summary>
		[Flags]
		public enum FrameProps
		{
			/// <summary>
			/// The width of a frame, in pixels.
			/// </summary>
			Width = 1,

			/// <summary>
			/// The height of a frame, in pixels.
			/// </summary>
			Height = 2,

			/// <summary>
			/// The stride of a row of pixels, in bytes. Only valid for <see cref="BitmapFrame"/>.
			/// </summary>
			Stride = 4,

			/// <summary>
			/// The base data type of pixels.
			/// </summary>
			Type = 8,

			/// <summary>
			/// The layout of colour components within a pixel.
			/// </summary>
			Layout = 16,

			/// <summary>
			/// The sequence number of a frame.
			/// </summary>
			SequenceNumber = 32,

			/// <summary>
			/// The pixel format of a frame.
			/// </summary>
			Format = FrameProps.Type | FrameProps.Layout,

			/// <summary>
			/// The sequence number of a frame. Same as FrameProps.Sequencenumber.
			/// </summary>
			Number = FrameProps.SequenceNumber,

			/// <summary>
			/// The size of a frame, i.e. width and height, in pixels.
			/// </summary>
			Size = FrameProps.Width | FrameProps.Height,
		}

		/// <summary>
		/// Asserts certain frame properties and throws an exception if they are violated.
		/// </summary>
		/// <typeparam name="TFrame">The type of the frame.</typeparam>
		/// <param name="bitmapFrames">A dictionary of BitmapFrames, indexed by pin name.</param>
		/// <param name="frameProps">The frame properties to check.</param>
		public static void AssertFrameProperties<TFrame>(Dictionary<string, TFrame> bitmapFrames, FrameProps frameProps) where TFrame : VideoFrame
		{
			TFrame frameA = null;
			TFrame frameB = null;

			// compare every two consecutive frames in the dictionary
			for (int a = 0, b = 1; a < b && b < bitmapFrames.Count; a++, b++)
			{
				if (frameA == null) frameA = bitmapFrames.Values.ElementAt(a);
				if (frameB == null) frameB = bitmapFrames.Values.ElementAt(b);

				// assert same width --------------------------------------------------------------
				if ((frameProps & FrameProps.Width) == FrameProps.Width)
				{
					if (frameA.Width != frameB.Width)
					{
						throw new Exception(String.Format(
							"Frames \"{0}\" ({1} x {2}) and \"{3}\" ({4} x {5}) need to have the same width.",
							bitmapFrames.First(x => x.Value == frameA).Key, frameA.Width, frameA.Height,
							bitmapFrames.First(x => x.Value == frameB).Key, frameB.Width, frameB.Height));
					}
				}

				// assert same height -------------------------------------------------------------------------
				if ((frameProps & FrameProps.Height) == FrameProps.Height)
				{
					if (frameA.Height != frameB.Height)
					{
						throw new Exception(String.Format(
							"Frames \"{0}\" ({1} x {2}) and \"{3}\" ({4} x {5}) need to have the same height.",
							bitmapFrames.First(x => x.Value == frameA).Key, frameA.Width, frameA.Height,
							bitmapFrames.First(x => x.Value == frameB).Key, frameB.Width, frameB.Height));
					}
				}

				// assert same stride -------------------------------------------------------------
				if ((frameProps & FrameProps.Stride) == FrameProps.Stride)
				{
					// stride is only checked for BitmapFrames
					var bitmapFrameA = frameA as BitmapFrame;
					var bitmapFrameB = frameB as BitmapFrame;

					if (bitmapFrameA != null && bitmapFrameB != null)
					{
						if (bitmapFrameA is BitmapFrame && bitmapFrameB is BitmapFrame)
						{
							if (bitmapFrameA.Stride != bitmapFrameB.Stride)
							{
								throw new Exception(String.Format(
									"Frames \"{0}\" ({1} x {2}, stride {3}) and \"{4}\" ({5} x {6}, stride {7}) need to have the same stride.",
									bitmapFrames.First(x => x.Value == bitmapFrameA).Key, bitmapFrameA.Width, bitmapFrameA.Height, bitmapFrameA.Stride,
									bitmapFrames.First(x => x.Value == bitmapFrameB).Key, bitmapFrameB.Width, bitmapFrameB.Height, bitmapFrameB.Stride));
							}
						}
					}
				}

				// assert same data type -----------------------------------------------------------
				if ((frameProps & FrameProps.Type) == FrameProps.Type)
				{
					if ((frameA.Format & PixelFormat.TypeMask) != (frameB.Format & PixelFormat.TypeMask))
					{
						throw new Exception(String.Format(
							"Frames \"{0}\" ({1}) and \"{2}\" ({3}) need to have the same data type of pixels.",
							bitmapFrames.First(x => x.Value == frameA).Key, frameA.Format,
							bitmapFrames.First(x => x.Value == frameB).Key, frameB.Format));
					}
				}

				// assert same colour layout ------------------------------------------------------
				if ((frameProps & FrameProps.Layout) == FrameProps.Layout)
				{
					if ((frameA.Format & PixelFormat.LayoutMask) != (frameB.Format & PixelFormat.LayoutMask))
					{
						throw new Exception(String.Format(
							"Frames \"{0}\" ({1}) and \"{2}\" ({3}) need to have the same layout of colour components.",
							bitmapFrames.First(x => x.Value == frameA).Key, frameA.Format,
							bitmapFrames.First(x => x.Value == frameB).Key, frameB.Format));
					}
				}

				// assert same sequence number ----------------------------------------------------
				if ((frameProps & FrameProps.SequenceNumber) == FrameProps.SequenceNumber)
				{
					if (frameA.SequenceNumber != frameB.SequenceNumber)
					{
						throw new Exception(String.Format(
							"Frames \"{0}\" (#{1}) and \"{2}\" (#{3}) need to have the same sequence number.",
							bitmapFrames.First(x => x.Value == frameA).Key, frameA.SequenceNumber,
							bitmapFrames.First(x => x.Value == frameB).Key, frameB.SequenceNumber));
					}
				}
			}
		}

		/// <summary>
		/// Asserts the pixel format of the given <see cref="VideoFrame"/> and throws an exception if it is wrong.
		/// </summary>
		/// <param name="frame">A <see cref="VideoFrame"/>.</param>
		/// <param name="pixelFormat">The desired pixel format.</param>
		public static void AssertPixelFormat(VideoFrame frame, PixelFormat pixelFormat)
		{
			if (frame.Format != pixelFormat)
			{
				throw new Exception(String.Format(
					"Frame with pixel format {0} needs to have the pixel format {1}.",
					frame.Format, pixelFormat));
			}
		}

		/// <summary>
		/// Asserts the pixel format of the given frames and throws an exception if they are wrong.
		/// </summary>
		/// <param name="frames">A dictionary of frames, indexed by pin name.</param>
		/// <param name="pixelFormat">The desired pixel format.</param>
		public static void AssertPixelFormat<TFrame>(Dictionary<string, TFrame> frames, PixelFormat pixelFormat) where TFrame : VideoFrame
		{
			foreach (var frame in frames)
			{
				// code is replicated here, so that the exception can say which pin's frame is at fault
				if (frame.Value.Format != pixelFormat)
				{
					throw new Exception(String.Format(
						"Frame \"{0}\" (pixel format: {1}) needs to have the pixel format {2}.",
						frame.Key, frame.Value.Format, pixelFormat));
				}
			}
		}

		//===============================================================================================
		#endregion Frame handling code
	}
}
