﻿// $Id: ExtrinsicParameters.cs 295 2010-08-05 13:08:16Z cr333 $

using System;
namespace VideoLib.Vision
{
	/// <summary>Collection of extrinsic camera parameters.</summary>
	public class ExtrinsicParameters
	{
		/// <summary>The 3-by-4 projection matrix encapsulating the extrinsic parameters.</summary>
		public float[] P { get; set; }

		/// <summary>Creates extrinsic parameters with identity rotation matrix and zero displacement.</summary>
		public ExtrinsicParameters()
		{
			P = new float[12];

			// identity rotation matrix
			P[ 0] = 1;
			P[ 5] = 1;
			P[10] = 1;
		}

		/// <summary>Gets or sets the <see cref="System.Single"/> at the specified index of the projection matrix.</summary>
		public float this[int index]
		{
			get
			{
				if (P == null || P.Length != 12) P = new float[12];
				return P[index];
			}
			set
			{
				if (P == null || P.Length != 12) P = new float[12];
				if (index >= 0 && index < P.Length) P[index] = value;
			}
		}

		/// <summary>Gets or sets the 3-by-3 rotation matrix part of the projection matrix.</summary>
		public float[] R
		{
			get
			{
				return new float[] { P[0], P[1], P[2], P[4], P[5], P[6], P[8], P[9], P[10] };
			}
			set
			{
				if (value.Length != 9) throw new ArgumentException("Rotation matrix requries 9 entries (3-by-3).");
				if (P == null || P.Length != 12) P = new float[12];

				P[0] = value[0]; P[1] = value[1]; P[ 2] = value[2];
				P[4] = value[3]; P[5] = value[4]; P[ 6] = value[5];
				P[8] = value[6]; P[9] = value[7]; P[10] = value[8];
			}
		}

		/// <summary>Gets or sets the 3-component translation vector part of the projection matrix.</summary>
		public float[] T
		{
			get
			{
				return new float[] { P[3], P[7], P[11] };
			}
			set
			{
				if (value.Length != 3) throw new ArgumentException("Translation vector requries 3 entries.");
				if (P == null || P.Length != 12) P = new float[12];

				P[ 3] = value[0];
				P[ 7] = value[1];
				P[11] = value[2];
			}
		}

		/// <summary>Sets the orientation using a 3-component rotation vector (according to Rodrigues formula).</summary>
		public float[] O
		{
			set
			{
				if (value.Length != 3) throw new ArgumentException("Rotation vector requries 3 entries.");
				if (P == null || P.Length != 12) P = new float[12];

				float theta = (float)Math.Sqrt(value[0] * value[0] + value[1] * value[1] + value[2] * value[2]);

				if(theta < Single.Epsilon)
				{
					R = new float[] { 1, 0, 0, 0, 1, 0, 0, 0, 1 };
				}
				else
				{
					float o1 = value[0] / theta;
					float o2 = value[1] / theta;
					float o3 = value[2] / theta;

					float a = (float)Math.Cos(theta);
					float b = (float)Math.Sin(theta);
					float c = 1 - a;

					P[ 0] =       a + o1 * o1 * c;
					P[ 1] = -o3 * b + o1 * o2 * c;
					P[ 2] =  o2 * b + o1 * o3 * c;
					P[ 4] =  o3 * b + o2 * o1 * c;
					P[ 5] =       a + o2 * o2 * c;
					P[ 6] = -o1 * b + o2 * o3 * c;
					P[ 8] = -o2 * b + o3 * o1 * c;
					P[ 9] =  o1 * b + o3 * o2 * c;
					P[10] =       a + o3 * o3 * c;
				}
			}
		}
	}
}